from __future__ import annotations

import logging
from typing import Any

import mne
import numpy as np

from ..base import BaseSolver, InverseOperator, SolverMeta

logger = logging.getLogger(__name__)


class SolverSESAME(BaseSolver):
    """SESAME-style dipole localization via Sequential Monte Carlo (SMC).

    This is a lightweight, discrete-grid implementation inspired by SESAME:
    it approximates the posterior over a small set of dipole locations using
    sequential importance sampling with resampling. Dipole positions are
    restricted to the vertices in the provided source space.

    Notes
    -----
    - The algorithm is stochastic; use `random_state` for reproducibility.
    - The forward model is converted to fixed orientation by BaseSolver.
    - The output is sparse with `n_dipoles` active vertices.
    """

    meta = SolverMeta(
        acronym="SESAME",
        full_name="Sequential Semi-Analytic Monte Carlo Estimation",
        category="Dipole Fitting",
        description=(
            "SESAME-style sequential Monte Carlo dipole fitting on a discrete source grid."
        ),
        references=[
            "Sommariva, S., & Sorrentino, A. (2014). Sequential Monte Carlo samplers for semi-linear inverse problems and application to magnetoencephalography. Inverse Problems, 30(11), 114020.",
        ],
    )

    def __init__(self, name: str = "SESAME", **kwargs: Any) -> None:
        self.name = name
        self.selected_dipoles: np.ndarray | None = None
        super().__init__(**kwargs)

    @staticmethod
    def _explained_energy_scores(leadfield: np.ndarray, data: np.ndarray) -> np.ndarray:
        norms = np.sum(leadfield * leadfield, axis=0)
        norms = np.where(norms <= 0, 1e-15, norms)
        proj = leadfield.T @ data
        return np.sum(proj * proj, axis=1) / norms

    def make_inverse_operator(  # type: ignore[override]
        self,
        forward,
        mne_obj,
        *args: Any,
        alpha: str | float = "auto",
        noise_cov: mne.Covariance | None = None,
        n: int | str | None = None,
        n_dipoles: int | str | None = None,
        n_particles: int = 64,
        n_candidates: int = 256,
        resample_threshold: float = 0.5,
        max_dipoles: int = 4,
        min_rel_rss_improvement: float = 0.02,
        noise_var: float | str = "auto",
        random_state: int | None = 0,
        **kwargs: Any,
    ):
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)

        data = self.unpack_data_obj(mne_obj)
        data = wf.sensor_transform @ data
        data = data - data.mean(axis=0, keepdims=True)

        leadfield = self.leadfield
        n_chans, n_sources = leadfield.shape

        # Decide number of dipoles to fit.
        force_n = n is not None or n_dipoles is not None
        if n_dipoles is None:
            n_dipoles = n
        if n_dipoles is None or (
            isinstance(n_dipoles, str) and n_dipoles.lower() == "auto"
        ):
            # Conservative default for benchmarks: estimate and cap.
            n_est = int(self.estimate_n_sources(data, method="enhanced"))
            n_dipoles_int = int(np.clip(n_est, 1, int(max_dipoles)))
        else:
            n_dipoles_int = max(int(n_dipoles), 1)
        n_dipoles_int = min(n_dipoles_int, n_sources)

        if isinstance(noise_var, str) and noise_var == "auto":
            # Start with per-entry variance; refined per-iteration using RSS spread.
            noise_var_val = float(np.var(data))
        else:
            noise_var_val = float(noise_var)
        noise_var_val = max(noise_var_val, 1e-15)

        rng = np.random.RandomState(random_state)

        particles: list[np.ndarray] = [
            np.array([], dtype=int) for _ in range(int(n_particles))
        ]
        weights = np.full(len(particles), 1.0 / len(particles), dtype=np.float64)

        residual_best = data
        best_rss = float(np.sum(residual_best * residual_best))
        best_dipoles = np.array([], dtype=int)

        for _k in range(int(n_dipoles_int)):
            scores = self._explained_energy_scores(leadfield, residual_best)
            n_candidates_eff = int(np.clip(n_candidates, 8, n_sources))
            cand = np.argsort(scores)[-n_candidates_eff:]
            cand_scores = np.maximum(scores[cand], 0.0)
            if float(cand_scores.sum()) <= 0:
                cand_scores = np.ones_like(cand_scores)
            cand_p = cand_scores / float(cand_scores.sum())

            new_particles: list[np.ndarray] = []
            residuals: list[np.ndarray] = []
            rss_list = np.zeros(len(particles), dtype=np.float64)

            # Always include the MAP proposal in the particle set for stability.
            best_cand = int(cand[int(np.argmax(scores[cand]))])

            for i in range(len(particles)):
                existing = set(particles[i].tolist())
                if i == 0 and best_cand not in existing:
                    new_idx = best_cand
                else:
                    new_idx = None
                    for _attempt in range(16):
                        proposal = int(rng.choice(cand, p=cand_p))
                        if proposal not in existing:
                            new_idx = proposal
                            break
                if new_idx is None:
                    for proposal in cand[::-1]:
                        if int(proposal) not in existing:
                            new_idx = int(proposal)
                            break
                if new_idx is None:
                    new_idx = int(np.argmax(scores))

                dipoles = np.append(particles[i], new_idx).astype(int)
                A = leadfield[:, dipoles]
                G = A.T @ A
                reg = 1e-12 * np.eye(G.shape[0])
                moments = np.linalg.solve(G + reg, A.T @ data)
                resid = data - A @ moments

                residuals.append(resid)
                new_particles.append(dipoles)

                rss_list[i] = float(np.sum(resid * resid))

            rss_min = float(np.min(rss_list))
            if not force_n:
                # Stop early if the next dipole doesn't improve the fit enough.
                rel_improve = (best_rss - rss_min) / max(best_rss, 1e-15)
                if rel_improve < float(min_rel_rss_improvement):
                    break

            # Convert RSS to stable log-weights (scale-free).
            rss_sorted = np.sort(rss_list)
            rss_med = float(rss_sorted[len(rss_sorted) // 2])
            temp = max(
                rss_med - rss_min, noise_var_val * n_chans * data.shape[1] * 1e-6, 1e-12
            )
            loglikes = -(rss_list - rss_min) / temp

            logw = np.log(weights + 1e-300) + loglikes
            logw -= float(np.max(logw))
            weights = np.exp(logw)
            weights /= float(np.sum(weights))

            particles = new_particles

            ess = 1.0 / float(np.sum(weights * weights))
            if ess < float(resample_threshold) * len(particles):
                idx = rng.choice(
                    np.arange(len(particles)), size=len(particles), p=weights
                )
                particles = [particles[int(j)].copy() for j in idx]
                residuals = [residuals[int(j)].copy() for j in idx]
                weights = np.full(
                    len(particles), 1.0 / len(particles), dtype=np.float64
                )

            best_idx = int(np.argmin(rss_list))
            residual_best = residuals[best_idx]
            best_dipoles = particles[best_idx].astype(int)
            best_rss = float(rss_list[best_idx])

        if best_dipoles.size == 0:
            # Fallback: match ECD (single best dipole)
            scores = self._explained_energy_scores(leadfield, data)
            best_dipoles = np.array([int(np.argmax(scores))], dtype=int)

        A = leadfield[:, best_dipoles]
        G = A.T @ A
        reg = 1e-12 * np.eye(G.shape[0])
        pinv = np.linalg.solve(G + reg, A.T)  # (k, n_chans)

        kernel = np.zeros((n_sources, n_chans), dtype=np.float64)
        for row, src_idx in enumerate(best_dipoles.tolist()):
            kernel[int(src_idx), :] += pinv[row, :]

        self.selected_dipoles = best_dipoles
        self.inverse_operators = [
            InverseOperator(kernel @ wf.sensor_transform, self.name)
        ]
        return self



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Any

import mne
import numpy as np

_TORCH_IMPORT_ERROR: ModuleNotFoundError | None = None
try:
    import torch  # type: ignore[import-not-found]
    import torch.nn.functional as F  # type: ignore[import-not-found]
except ModuleNotFoundError as exc:  # pragma: no cover
    torch = None  # type: ignore[assignment]
    F = None  # type: ignore[assignment]
    _TORCH_IMPORT_ERROR = exc

from sklearn.covariance import OAS

from ...simulate import SimulationGenerator
from ..base import BaseSolver, InverseOperator, SolverMeta
from .covcnn import _CovCNNNet
from .torch_utils import count_trainable_parameters, get_torch_device

logger = logging.getLogger(__name__)


class SolverCovCNNKL(BaseSolver):
    """CovCNN variant trained with KL divergence on L1-normalized source maps.

    This treats the collapsed source activity as a probability distribution
    (after optional power compression) and trains the network to predict a
    distribution (via softmax). It tends to penalize missing secondary sources
    more than cosine similarity on max-normalized maps.
    """

    meta = SolverMeta(
        acronym="CovCNN-KL",
        full_name="CovCNN (KL divergence)",
        category="Neural Networks",
        description=(
            "Supervised ANN on sensor covariance trained with KL divergence "
            "between predicted and true L1-normalized source distributions."
        ),
        references=["Lukas Hecker 2025, unpublished"],
    )

    def __init__(
        self,
        name: str = "CovCNN (KL)",
        *,
        reduce_rank: bool = False,
        use_shrinkage: bool = True,
        **kwargs,
    ) -> None:
        self.name = name
        self.use_shrinkage = bool(use_shrinkage)
        self.model: Any = None
        self.optimizer: Any = None
        self.device: Any = None
        self.generator: Any = None
        return super().__init__(reduce_rank=reduce_rank, **kwargs)

    def make_inverse_operator(  # type: ignore[override]
        self,
        forward,
        simulation_config,
        *args,
        n_dense_units: int = 500,
        n_dense_layers: int = 2,
        activation_function: str = "tanh",
        epochs: int = 1000,
        learning_rate: float = 1e-3,
        patience: int = 400,
        cov_type: str = "basic",
        target_power: float = 0.5,
        temperature: float = 1.0,
        gamma_power: float = 1.5,
        noise_cov: mne.Covariance | None = None,
        alpha: str | float = "auto",
        **kwargs,
    ):
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)
        self.forward = forward
        self.simulation_config = simulation_config

        self.n_dense_units = int(n_dense_units)
        self.n_dense_layers = int(n_dense_layers)
        self.activation_function = str(activation_function)
        self.output_activation = "linear"  # logits
        self.epochs = int(epochs)
        self.learning_rate = float(learning_rate)
        self.patience = int(patience)
        self.cov_type = str(cov_type)
        self.target_power = float(target_power)
        self.temperature = float(temperature)
        self.gamma_power = float(gamma_power)

        logger.info("Create generator…")
        self.create_generator()
        logger.info("Build model…")
        self.build_model()
        logger.info("Train model…")
        self.train_model()

        self.inverse_operators: list = []
        return self

    def apply_inverse_operator(self, mne_obj, prior=None) -> mne.SourceEstimate:
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data
        source_mat = self.apply_model(data, prior=prior)
        return self.source_to_object(source_mat)

    def _shrinkage_covariance(self, Y: np.ndarray) -> np.ndarray:
        lw = OAS(assume_centered=False)
        return lw.fit(Y.T).covariance_

    def compute_covariance(self, Y: np.ndarray) -> np.ndarray:
        if self.cov_type != "basic":
            raise ValueError(
                f"cov_type={self.cov_type!r} not supported in SolverCovCNNKL (use 'basic')."
            )
        C = Y @ Y.T
        if self.use_shrinkage:
            C = self._shrinkage_covariance(Y)
        return C

    def create_generator(self) -> None:
        sim_gen = SimulationGenerator(self.forward, config=self.simulation_config)

        def wrapped_generator():
            for x, y, _info in sim_gen.generate():
                x_cov = np.stack([self.compute_covariance(xx) for xx in x], axis=0)
                max_abs = np.abs(x_cov).max(axis=(1, 2), keepdims=True)
                max_abs = np.where(max_abs == 0, 1.0, max_abs)
                x_cov = (x_cov / max_abs).astype(np.float32, copy=False)
                x_cov = x_cov[:, np.newaxis, :, :]  # NCHW

                y_abs_mean = np.abs(y).mean(axis=2)
                if self.target_power != 1.0:
                    y_abs_mean = y_abs_mean ** float(self.target_power)
                y_sum = y_abs_mean.sum(axis=1, keepdims=True)
                y_sum = np.where(y_sum == 0, 1.0, y_sum)
                y_dist = (y_abs_mean / y_sum).astype(np.float32, copy=False)

                yield x_cov, y_dist

        self.generator = wrapped_generator()

    def build_model(self) -> None:
        if _TORCH_IMPORT_ERROR is not None:  # pragma: no cover
            raise ImportError(
                "PyTorch is required for neural-network solvers. "
                'Install it via `pip install "invertmeeg[ann]"` (or install `torch` directly).'
            ) from _TORCH_IMPORT_ERROR

        self.device = get_torch_device()
        self.model = _CovCNNNet(
            self.leadfield,
            n_dense_layers=int(self.n_dense_layers),
            n_dense_units=int(self.n_dense_units),
            activation_function=str(self.activation_function),
            output_activation=str(self.output_activation),
        ).to(self.device)
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=float(self.learning_rate)
        )
        logger.info(
            "Total number of trainable parameters: %d",
            count_trainable_parameters(self.model),
        )

    def train_model(self) -> None:
        if self.model is None:
            raise RuntimeError("Model not initialized. Call build_model() first.")
        if self.optimizer is None:
            raise RuntimeError("Optimizer not initialized. Call build_model() first.")

        # Re-initialize generator.
        self.create_generator()

        # Validation batch.
        x_val, y_val = next(self.generator)
        device = self.device or get_torch_device()
        x_val_t = torch.as_tensor(x_val, dtype=torch.float32, device=device)
        y_val_t = torch.as_tensor(y_val, dtype=torch.float32, device=device)

        best_val = float("inf")
        best_state = None
        patience_left = int(self.patience)
        log_every = 10

        for epoch in range(int(self.epochs)):
            self.model.train()
            x_batch, y_batch = next(self.generator)
            x_t = torch.as_tensor(x_batch, dtype=torch.float32, device=device)
            y_t = torch.as_tensor(y_batch, dtype=torch.float32, device=device)

            self.optimizer.zero_grad(set_to_none=True)
            logits = self.model(x_t) / float(self.temperature)
            log_probs = F.log_softmax(logits, dim=-1)
            loss = F.kl_div(log_probs, y_t, reduction="batchmean")
            loss.backward()
            self.optimizer.step()

            self.model.eval()
            with torch.no_grad():
                v_logits = self.model(x_val_t) / float(self.temperature)
                v_log_probs = F.log_softmax(v_logits, dim=-1)
                val_loss = float(
                    F.kl_div(v_log_probs, y_val_t, reduction="batchmean").cpu().item()
                )

            if val_loss < best_val:
                best_val = val_loss
                best_state = deepcopy(self.model.state_dict())
                patience_left = int(self.patience)
                logger.info(
                    "Epoch %d/%d - loss=%.6f val_loss=%.6f (new best)",
                    epoch + 1,
                    int(self.epochs),
                    float(loss.detach().cpu().item()),
                    val_loss,
                )
            else:
                patience_left -= 1
                if (epoch == 0) or ((epoch + 1) % log_every == 0):
                    logger.info(
                        "Epoch %d/%d - loss=%.6f val_loss=%.6f (patience_left=%d)",
                        epoch + 1,
                        int(self.epochs),
                        float(loss.detach().cpu().item()),
                        val_loss,
                        patience_left,
                    )
                if patience_left <= 0:
                    logger.info(
                        "Early stopping at epoch %d/%d (best_val=%.6f)",
                        epoch + 1,
                        int(self.epochs),
                        best_val,
                    )
                    break

        if best_state is not None:
            self.model.load_state_dict(best_state)
        self.model.eval()

    def apply_model(self, data: np.ndarray, prior=None) -> np.ndarray:
        y = deepcopy(data)
        y = y - y.mean(axis=1, keepdims=True)
        n_channels, _n_times = y.shape

        C = self.compute_covariance(y)
        max_abs = float(np.abs(C).max())
        if max_abs > 0:
            C = C / max_abs
        C = C[np.newaxis, np.newaxis, :, :].astype(np.float32, copy=False)

        assert self.model is not None
        self.model.eval()
        device = self.device or get_torch_device()
        with torch.no_grad():
            logits = self.model(
                torch.as_tensor(C, dtype=torch.float32, device=device)
            ) / float(self.temperature)
            probs = torch.softmax(logits, dim=-1).detach().cpu().numpy()[0]

        max_p = float(np.max(probs))
        gammas = probs / max_p if max_p > 0 else probs
        if getattr(self, "gamma_power", 1.0) != 1.0:
            gammas = gammas ** float(self.gamma_power)
            max_gamma = float(np.max(gammas))
            if max_gamma > 0:
                gammas = gammas / max_gamma

        if prior is not None:
            prior = np.asarray(prior, dtype=float)
            prior_max = float(np.max(prior))
            if prior_max > 0:
                gammas = gammas * (prior / prior_max)

        source_covariance = np.diag(gammas.astype(np.float64, copy=False))

        Sigma_y = self.leadfield @ source_covariance @ self.leadfield.T
        if self.alpha == "auto":
            r_grid = np.asarray(self.r_values, dtype=float)
        else:
            r_grid = np.asarray([float(self.alpha)], dtype=float)
        self.alphas = list(r_grid)

        inverse_ops = []
        trace_Sy = float(np.trace(Sigma_y))
        if not np.isfinite(trace_Sy) or trace_Sy <= 0:
            trace_Sy = 1.0
        for r in r_grid:
            reg_term = float(r) * trace_Sy / float(n_channels)
            inv = np.linalg.inv(Sigma_y + reg_term * np.eye(n_channels))
            W = source_covariance @ self.leadfield.T @ inv
            inverse_ops.append(W)

        self.inverse_operators = [InverseOperator(op, self.name) for op in inverse_ops]
        x_hat, _ = self.regularise_gcv(y)
        return x_hat



import mne
import numpy as np

from ..base import BaseSolver, InverseOperator, SolverMeta


class SolverEPIFOCUS(BaseSolver):
    """Class for the EPIFOCUS inverse solution [1].

    EPIFOCUS normalizes leadfield columns to unit norm before computing the
    inverse, then divides the result by the original column norms.  This
    promotes focal solutions because sources with small leadfield norms (deep
    sources) receive a larger weight in the final estimate.

    References
    ----------
    [1] Menendez, R. G. D. P., Andino, S. G., Lantz, G., Michel, C. M., &
    Landis, T. (2001). Noninvasive localization of electromagnetic epileptic
    activity. I. Method descriptions and simulations. Brain topography, 14(2),
    131-137.

    """

    meta = SolverMeta(
        acronym="EPIFOCUS",
        full_name="EPIFOCUS",
        category="Minimum Norm",
        description=(
            "Epileptic focus localization method that uses source-wise "
            "normalization/weighting to promote focal solutions."
        ),
        references=[
            "Grave de Peralta Menendez, R., Gonzalez Andino, S., Lantz, G., Michel, C. M., & Landis, T. (2001). Noninvasive localization of electromagnetic epileptic activity. I. Method descriptions and simulations. Brain Topography, 14(2), 131–137.",
        ],
    )
    SUPPORTS_VECTOR_ORIENTATION: bool = True

    def __init__(self, name="EPIFOCUS", eps=1e-15, **kwargs):
        self.name = name
        self.eps = float(eps)
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)
        leadfield = wf.G_white

        n_chans, n_dipoles = leadfield.shape

        # Column normalization: normalize each source's leadfield to unit norm
        col_norms = np.linalg.norm(leadfield, axis=0)
        col_norms = np.maximum(col_norms, self.eps)
        L_norm = leadfield / col_norms

        # Compute regularized inverse of the normalized leadfield.
        # The 1/col_norms weighting promotes focal (deep) sources.
        LLT = L_norm @ L_norm.T
        self.get_alphas(reference=LLT)

        inverse_operators = []
        for alpha in self.alphas:
            K = np.linalg.solve(LLT + float(alpha) * np.identity(n_chans), L_norm).T
            # Denormalize: divide by col_norms to weight toward focal sources
            K = (1.0 / col_norms[:, np.newaxis]) * K
            inverse_operators.append(K @ wf.sensor_transform)

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]
        return self



"""
MUSIC / Subspace solvers for M/EEG source reconstruction.
"""

from .adaptive_alternating_projections import SolverAdaptiveAlternatingProjections
from .alternating_projections import SolverAlternatingProjections
from .exhaustive_subspace_ml import SolverExhaustiveSubspaceML
from .exso_music import SolverExSoMUSIC
from .flex_greedy_ml import SolverFlexGreedyML
from .flex_music import SolverFLEXMUSIC
from .flex_music_2 import SolverFLEXMUSIC_2
from .generalized_iterative import SolverGeneralizedIterative
from .greedy_ml import SolverGreedyML
from .music import SolverMUSIC
from .signal_subspace_matching import SolverSignalSubspaceMatching

__all__ = [
    "SolverMUSIC",
    "SolverExSoMUSIC",
    "SolverFLEXMUSIC",
    "SolverSignalSubspaceMatching",
    "SolverAlternatingProjections",
    "SolverAdaptiveAlternatingProjections",
    "SolverFLEXMUSIC_2",
    "SolverGeneralizedIterative",
    "SolverGreedyML",
    "SolverExhaustiveSubspaceML",
    "SolverFlexGreedyML",
]



import logging

import mne
import numpy as np

logger = logging.getLogger(__name__)


class ASR:
    """
    Artifact Subspace Reconstruction (ASR) implementation.

    Parameters
    ----------
    sfreq : float
        Sampling frequency (Hz). Required if input is np.ndarray.
    window_size : float
        Window length in seconds (default 0.5).
    step_size : float
        Step size as fraction of window (default 0.5 → 50% overlap).
    cutoff : float
        Standard deviation threshold (default 5.0).
    """

    def __init__(self, sfreq=None, window_size=0.5, step_size=0.5, cutoff=5.0):
        self.sfreq = sfreq
        self.window_size = window_size
        self.step_size = step_size
        self.cutoff = cutoff
        self.V = None
        self.lam_ref = None

    # -------------------------- core utilities --------------------------

    @staticmethod
    def _window_cov(X, win_len, step):
        """Compute covariance matrices from sliding windows."""
        n_ch, n_samp = X.shape
        step_len = int(win_len * (1 - step))
        covs = []
        for start in range(0, n_samp - win_len + 1, step_len):
            seg = X[:, start : start + win_len]
            seg -= seg.mean(axis=1, keepdims=True)
            covs.append(np.cov(seg))
        return np.array(covs)

    def _fit_reference(self, X):
        """Fit reference covariance and thresholds."""
        win_len = int(self.window_size * self.sfreq)
        covs = self._window_cov(X, win_len, self.step_size)
        C_ref = covs.mean(axis=0)
        self.V, lam, _ = np.linalg.svd(C_ref)
        log_lam = np.log10(lam + 1e-12)
        self.lam_ref = lam * 10 ** (self.cutoff * log_lam.std())
        return C_ref

    # -------------------------- main logic --------------------------

    def run(self, data: np.ndarray, calibration=None) -> np.ndarray:
        """
        Run ASR on numpy array data.

        Parameters
        ----------
        data : np.ndarray, shape (n_channels, n_times)
        calibration : np.ndarray or None
            Optional baseline data for reference fit.

        Returns
        -------
        cleaned : np.ndarray
            Denoised signal (same shape as input).
        """
        X = np.asarray(data)
        n_ch, n_samp = X.shape

        if self.sfreq is None:
            raise ValueError("Sampling frequency (sfreq) must be provided.")

        # For very short recordings, skip ASR
        win_len = int(self.window_size * self.sfreq)
        if n_samp < win_len * 2:
            logger.warning("[ASR] Too few samples for stable covariance estimation.")
            return X.copy()

        # Fit reference if not done
        if calibration is None:
            calibration = X
        self._fit_reference(calibration)

        # Process in windows
        step_len = int(win_len * (1 - self.step_size))
        cleaned = np.zeros_like(X)
        weights = np.zeros(n_samp)

        for start in range(0, n_samp - win_len + 1, step_len):
            seg = X[:, start : start + win_len]
            seg -= seg.mean(axis=1, keepdims=True)
            C = np.cov(seg)
            z = np.diag(self.V.T @ C @ self.V)
            z_clipped = np.minimum(z, self.lam_ref)
            C_clipped = self.V @ np.diag(z_clipped) @ self.V.T

            # Whitening and projection back
            eigvals, eigvecs = np.linalg.eigh(C)
            W = eigvecs @ np.diag(1.0 / np.sqrt(eigvals + 1e-12)) @ eigvecs.T
            S = np.linalg.cholesky(C_clipped + 1e-12 * np.eye(n_ch))
            seg_clean = S @ W @ seg

            cleaned[:, start : start + win_len] += seg_clean
            weights[start : start + win_len] += 1.0

        cleaned /= np.maximum(weights, 1e-12)
        return cleaned

    # -------------------------- mne wrapper --------------------------

    def run_mne(self, mne_obj: mne.io.BaseRaw | mne.Epochs | mne.Evoked):
        """
        Run ASR on an MNE object and return a cleaned copy.
        """
        if isinstance(mne_obj, mne.io.BaseRaw):
            data = mne_obj.get_data()
            self.sfreq = mne_obj.info["sfreq"]
            cleaned = self.run(data)
            new_raw = mne.io.RawArray(cleaned, mne_obj.info)
            return new_raw

        elif isinstance(mne_obj, mne.Epochs):
            self.sfreq = mne_obj.info["sfreq"]
            cleaned_data = []
            for ep in mne_obj.get_data():
                cleaned_data.append(self.run(ep.T).T)
            new_epochs = mne.EpochsArray(
                np.array(cleaned_data),
                mne_obj.info,
                events=mne_obj.events,
                tmin=mne_obj.tmin,
            )
            return new_epochs

        elif isinstance(mne_obj, mne.Evoked):
            self.sfreq = mne_obj.info["sfreq"]
            cleaned = self.run(mne_obj.data)
            return mne.EvokedArray(cleaned, mne_obj.info, tmin=mne_obj.times[0])

        else:
            raise TypeError(f"Unsupported MNE object type: {type(mne_obj)}")



import logging

import mne
import numpy as np

from ...util import (
    build_source_adjacency,
    thresholding,
)
from ..base import BaseSolver, SolverMeta

logger = logging.getLogger(__name__)


class SolverSP(BaseSolver):
    """Subspace Pursuit (SP) with MMV and patch-dictionary support.

    Greedy sparse recovery that alternates between support expansion and
    pruning, solving a least-squares fit on the candidate support each
    iteration.  All time points are used jointly (MMV) and spatial patches
    can be selected via an adjacency-based smooth dictionary.

    References
    ----------
    [1] Dai, W., & Milenkovic, O. (2009). Subspace pursuit for compressive
    sensing signal reconstruction. IEEE Transactions on Information Theory,
    55(5), 2230-2249.

    [2] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on Signal Processing, 59(9),
    4053-4085.
    """

    meta = SolverMeta(
        acronym="SP",
        full_name="Subspace Pursuit",
        category="Matching Pursuit",
        description=(
            "Greedy sparse recovery that alternates between support expansion and "
            "pruning (singletons or patches) using all time points jointly (MMV), "
            "solving a least-squares fit on the candidate support each iteration."
        ),
        references=[
            "Dai, W., & Milenkovic, O. (2009). Subspace pursuit for compressive sensing signal reconstruction. IEEE Transactions on Information Theory, 55(5), 2230–2249.",
            "Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing: From theory to applications. IEEE Transactions on Signal Processing, 59(9), 4053–4085.",
        ],
    )

    def __init__(self, name="Subspace Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)

        # Adjacency and patch dictionary
        adjacency = build_source_adjacency(self.forward["src"], verbose=0).toarray()
        self.adjacency = adjacency
        patch_operator = adjacency + np.eye(adjacency.shape[0])

        self.leadfield_original = self.leadfield.copy()
        leadfield_smooth = self.leadfield_original @ patch_operator

        self.leadfield_smooth_normed = self.robust_normalize_leadfield(leadfield_smooth)
        self.leadfield_normed = self.robust_normalize_leadfield(self.leadfield_original)

        self.inverse_operators = []
        return self

    def apply_inverse_operator(
        self,
        mne_obj,
        K="auto",
        include_singletons=True,
        include_patches=False,
    ) -> mne.SourceEstimate:
        """Apply the SP inverse solution.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        K : int | "auto"
            The sparsity level (number of atoms to keep after pruning).
        include_singletons : bool
            Include individual dipoles as candidate atoms.
        include_patches : bool
            Include spatial-patch atoms built from source adjacency.

        Return
        ------
        stc : mne.SourceEstimate
            The source estimate containing the inverse solution.
        """
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data
        source_mat = self.calc_sp_solution(
            data,
            K=K,
            include_singletons=include_singletons,
            include_patches=include_patches,
        )
        stc = self.source_to_object(source_mat)
        return stc

    def _select_atoms(self, R, K, include_singletons, include_patches):
        """Select K candidate atoms from residual using aggregated correlations."""
        if include_patches and include_singletons:
            b_smooth = np.linalg.norm(self.leadfield_smooth_normed.T @ R, axis=1)
            b_sparse = np.linalg.norm(self.leadfield_normed.T @ R, axis=1)
            if b_sparse.max() > b_smooth.max():
                b_thresh = thresholding(b_sparse, K)
                return np.where(b_thresh != 0)[0]
            else:
                b_thresh = thresholding(b_smooth, K)
                best_idx = np.where(b_thresh != 0)[0]
                patches = []
                for idx in best_idx:
                    patch = np.where(self.adjacency[idx] != 0)[0]
                    patch = np.append(patch, idx)
                    patches.append(patch)
                return (
                    np.unique(np.concatenate(patches))
                    if patches
                    else np.array([], dtype=int)
                )
        elif include_patches:
            b_smooth = np.linalg.norm(self.leadfield_smooth_normed.T @ R, axis=1)
            b_thresh = thresholding(b_smooth, K)
            best_idx = np.where(b_thresh != 0)[0]
            patches = []
            for idx in best_idx:
                patch = np.where(self.adjacency[idx] != 0)[0]
                patch = np.append(patch, idx)
                patches.append(patch)
            return (
                np.unique(np.concatenate(patches))
                if patches
                else np.array([], dtype=int)
            )
        else:  # include_singletons only
            b_sparse = np.linalg.norm(self.leadfield_normed.T @ R, axis=1)
            b_thresh = thresholding(b_sparse, K)
            return np.where(b_thresh != 0)[0]

    def calc_sp_solution(
        self, y, K="auto", include_singletons=True, include_patches=False
    ):
        """Calculates the Subspace Pursuit (SP) inverse solution (MMV, patch-aware).

        Parameters
        ----------
        y : numpy.ndarray
            The data matrix, shape ``(n_chans,)`` or ``(n_chans, n_time)``.
        K : int | "auto"
            Sparsity level.
        include_singletons : bool
            Include individual-dipole atoms.
        include_patches : bool
            Include spatial-patch atoms.

        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution.
        """
        if not include_singletons and not include_patches:
            raise ValueError(
                "At least one of include_patches/include_singletons must be True"
            )

        squeeze = False
        if y.ndim == 1:
            y = y[:, np.newaxis]
            squeeze = True

        n_chans, n_time = y.shape
        _, n_dipoles = self.leadfield_original.shape

        if K == "auto":
            K = min(8, max(2, n_chans // 10))
        K = int(K)
        if K <= 0:
            raise ValueError("K must be positive")

        # Initial atom selection
        T0 = self._select_atoms(y, K, include_singletons, include_patches)

        # Initial residual
        if len(T0) > 0:
            L_T0 = self.leadfield_original[:, T0]
            x_T0 = self.robust_inverse_solution(L_T0, y)
            R = y - L_T0 @ x_T0
        else:
            R = y.copy()

        T_list = [T0]
        R_list = [R]

        for i in range(1, n_chans + 1):
            # Select K new candidates from residual
            new_T = self._select_atoms(
                R_list[-1], K, include_singletons, include_patches
            )
            T_tilde = np.unique(np.concatenate([T_list[i - 1], new_T]))

            # Solve on expanded support and prune by coefficient norms
            if len(T_tilde) > 0:
                L_tilde = self.leadfield_original[:, T_tilde]
                xp = self.robust_inverse_solution(L_tilde, y)
                xp_norm = np.linalg.norm(xp, axis=1)
                T_l = T_tilde[np.where(thresholding(xp_norm, K) != 0)[0]]
            else:
                T_l = T_tilde

            T_list.append(T_l)

            # Calculate residual
            if len(T_l) > 0:
                L_final = self.leadfield_original[:, T_l]
                xp_final = self.robust_inverse_solution(L_final, y)
                R = y - L_final @ xp_final
            else:
                R = y.copy()

            R_list.append(R)

            # Early stopping
            if len(R_list) > 1 and (
                np.linalg.norm(R_list[-1]) >= np.linalg.norm(R_list[-2]) or i == n_chans
            ):
                T_l = T_list[-2]
                break
            else:
                T_l = T_list[-1]

        x_hat = np.zeros((n_dipoles, n_time))
        if len(T_l) > 0:
            L_final = self.leadfield_original[:, T_l]
            x_hat[T_l] = self.robust_inverse_solution(L_final, y)

        if squeeze:
            x_hat = x_hat[:, 0]
        return x_hat



"""Whitened ReciPSIICOS Beamformer."""

from __future__ import annotations

import mne
import numpy as np

from ..base import InverseOperator, SolverMeta
from .base_beamformer import BaseBeamformer
from .utils import (
    _cov,
    _lcmv_inverse_operator,
    _project_covariance_whitened,
    _psd_spectral_flip,
    _virtual_sensors,
)


class SolverReciPSIICOSWhitened(BaseBeamformer):
    """
    Whitened ReciPSIICOS Beamformer for M/EEG inverse solution (recommended).

    Extends plain ReciPSIICOS by additionally removing correlation subspace
    components in a whitened representation, providing better suppression of
    correlated source cross-talk.

    References
    ----------
    [1] Kuznetsova, A., Nurislamova, Y., & Ossadtchi, A. (2021). Modified
        covariance beamformer for solving MEG inverse problem in the environment
        with correlated sources. Neuroimage, 228, 117677.
    """

    SUPPORTS_VECTOR_ORIENTATION: bool = False

    meta = SolverMeta(
        slug="recipsiicos_whitened",
        full_name="ReciPSIICOS (whitened)",
        category="Beamformers",
        description=(
            "Whitened ReciPSIICOS variant that additionally removes correlation "
            "subspace components in a whitened representation."
        ),
        references=[
            "Kuznetsova, A., Nurislamova, Y., & Ossadtchi, A. (2021). Modified "
            "covariance beamformer for solving MEG inverse problem in the environment "
            "with correlated sources. NeuroImage, 228, 117677.",
        ],
    )

    def __init__(
        self, name="ReciPSIICOS-Whitened", reduce_rank=True, rank="auto", **kwargs
    ):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        virtual_sensor_energy=0.99,
        pwr_energy=0.9,
        pwr_rank=None,
        cor_rank=None,
        max_pairs=None,
        seed=0,
        spectral_flip=True,
        verbose=0,
        **kwargs,
    ):
        """
        Calculate inverse operator using whitened ReciPSIICOS method.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float or 'auto'
            The regularization parameter for LCMV backend.
        noise_cov : array, optional
            Noise covariance matrix for whitening.
        virtual_sensor_energy : float
            Fraction of leadfield energy to retain (default: 0.99)
        pwr_energy : float
            Fraction of power subspace energy to retain (default: 0.995)
        pwr_rank : int, optional
            Explicit power subspace rank (overrides pwr_energy)
        cor_rank : int, optional
            Number of correlation modes to remove (default: auto)
        max_pairs : int, optional
            Max source pairs to sample (default: all)
        seed : int
            Random seed for pair sampling (default: 0)
        spectral_flip : bool
            Apply spectral flip to ensure PSD (default: True)

        Returns
        -------
        self : object
            Returns itself for convenience
        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        # 1) Whiten leadfield and data via standard pipeline (SSP + noise whitening)
        wf = self.prepare_whitened_forward(noise_cov)
        col_norms = np.linalg.norm(wf.G_white, axis=0)
        col_norms = np.maximum(col_norms, 1e-30)
        L = wf.G_white / col_norms
        Yw = wf.sensor_transform @ data

        # 2) Virtual sensors via leadfield SVD
        UrT, Lr = _virtual_sensors(L, keep_energy=virtual_sensor_energy)
        Yr = UrT @ Yw

        # 3) Data covariance in reduced space
        Cr = _cov(Yr)

        # 4) Whitened ReciPSIICOS projection
        Ct = _project_covariance_whitened(
            Cr,
            Lr,
            pwr_energy=pwr_energy,
            pwr_rank=pwr_rank,
            cor_rank=cor_rank,
            max_pairs=max_pairs,
            seed=seed,
        )

        # 5) Spectral flip
        if spectral_flip:
            Ct = _psd_spectral_flip(Ct)

        # 6) Build inverse operators for different regularization values
        # ReciPSIICOS uses trace-shrinkage regularization: C + lam*trace(C)/m*I,
        # where lam is dimensionless. Do NOT eigenvalue-scale it.
        if alpha == "auto":
            reg_values = np.logspace(-6, 1, self.n_reg_params)
        else:
            reg_values = np.asarray([float(alpha)], dtype=float)
        self.alphas = list(np.asarray(reg_values, dtype=float))

        inverse_operators = []
        for reg in self.alphas:
            W_reduced = _lcmv_inverse_operator(Lr, Ct, reg=reg)
            # Map from reduced whitened space back to raw sensor space
            W_full = W_reduced @ UrT @ wf.sensor_transform
            inverse_operator = W_full
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [
            InverseOperator(inv_op, self.name) for inv_op in inverse_operators
        ]
        return self



import logging
from copy import deepcopy

import mne
import numpy as np

from ...util import (
    best_index_residual,
    build_source_adjacency,
    thresholding,
)
from ..base import BaseSolver, SolverMeta

logger = logging.getLogger(__name__)


class SolverOMP(BaseSolver):
    """Orthogonal Matching Pursuit (OMP) with MMV and patch-dictionary support.

    Greedy sparse recovery that iteratively selects the best-correlating atom
    (singleton dipole or spatial patch) and re-solves a least-squares fit on the
    selected support.  All time points are used jointly (MMV) for atom selection.

    References
    ----------
    [1] Tropp, J. A., & Gilbert, A. C. (2007). Signal recovery from random
    measurements via orthogonal matching pursuit. IEEE Transactions on
    Information Theory, 53(12), 4655-4666.

    [2] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on Signal Processing, 59(9),
    4053-4085.
    """

    meta = SolverMeta(
        acronym="OMP",
        full_name="Orthogonal Matching Pursuit",
        category="Matching Pursuit",
        description=(
            "Greedy sparse recovery that iteratively selects the best-correlating atom "
            "(singleton or spatial patch) using all time points jointly (MMV) and "
            "re-solves a least-squares fit on the selected support."
        ),
        references=[
            "Rezaiifar, R., & Krishnaprasad, P. S. (1995). Orthogonal matching pursuit: Recursive function approximation with applications to wavelet decomposition. Proceedings of the 27th Annual Asilomar Conference on Signals, Systems, and Computers.",
            "Tropp, J. A., & Gilbert, A. C. (2007). Signal recovery from random measurements via orthogonal matching pursuit. IEEE Transactions on Information Theory, 53(12), 4655–4666.",
        ],
    )

    def __init__(self, name="Orthogonal Matching Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)

        # Adjacency and patch dictionary
        adjacency = build_source_adjacency(self.forward["src"], verbose=0).toarray()
        self.adjacency = adjacency
        patch_operator = adjacency + np.eye(adjacency.shape[0])

        self.leadfield_original = self.leadfield.copy()
        leadfield_smooth = self.leadfield_original @ patch_operator

        self.leadfield_smooth_normed = self.robust_normalize_leadfield(leadfield_smooth)
        self.leadfield_normed = self.robust_normalize_leadfield(self.leadfield_original)

        self.inverse_operators = []
        return self

    def apply_inverse_operator(
        self,
        mne_obj,
        K="auto",
        max_iter=None,
        include_singletons=True,
        include_patches=False,
    ) -> mne.SourceEstimate:
        """Apply the inverse operator.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        K : int
            The number of atoms to select per iteration.
        max_iter : int
            The maximum number of iterations.
        include_singletons : bool
            Include individual dipoles as candidate atoms.
        include_patches : bool
            Include spatial-patch atoms built from source adjacency.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object
        """
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data
        source_mat = self.calc_omp_solution(
            data,
            K=K,
            max_iter=max_iter,
            include_singletons=include_singletons,
            include_patches=include_patches,
        )
        stc = self.source_to_object(source_mat)
        return stc

    def calc_omp_solution(
        self, y, K="auto", max_iter=None, include_singletons=True, include_patches=False
    ):
        """Calculates the OMP inverse solution (MMV, patch-aware).

        Parameters
        ----------
        y : numpy.ndarray
            The data matrix, shape ``(n_chans,)`` or ``(n_chans, n_time)``.
        K : int | "auto"
            The number of atoms to select per iteration.
        max_iter : int | None
            Maximum number of greedy iterations.
        include_singletons : bool
            Include individual-dipole atoms.
        include_patches : bool
            Include spatial-patch atoms.

        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution, shape ``(n_dipoles,)`` or ``(n_dipoles, n_time)``.
        """
        if not include_singletons and not include_patches:
            raise ValueError(
                "At least one of include_patches/include_singletons must be True"
            )

        # Handle both SMV and MMV input
        squeeze = False
        if y.ndim == 1:
            y = y[:, np.newaxis]
            squeeze = True

        n_chans, n_time = y.shape
        if K == "auto":
            K = 1
        K = int(K)
        if K <= 0:
            raise ValueError("K must be positive")

        if max_iter is None:
            max_iter = int(n_chans / 2)
        _, n_dipoles = self.leadfield_original.shape

        x_hat = np.zeros((n_dipoles, n_time))
        x_hats = [deepcopy(x_hat)]

        R = y.copy()
        omega = np.array([], dtype=int)

        y_hat = self.leadfield_original @ x_hat
        residuals = np.array([np.linalg.norm(y - y_hat)])

        for _ in range(max_iter):
            # Aggregated correlation across time (L2 norm over time dimension)
            if include_patches and include_singletons:
                b_smooth = np.linalg.norm(
                    self.leadfield_smooth_normed.T @ R, axis=1, ord=2
                )
                b_sparse = np.linalg.norm(self.leadfield_normed.T @ R, axis=1, ord=2)
                if b_sparse.max() > b_smooth.max():
                    b_thresh = thresholding(b_sparse, K)
                    new_atoms = np.where(b_thresh != 0)[0]
                else:
                    b_thresh = thresholding(b_smooth, K)
                    best_idx = np.where(b_thresh != 0)[0]
                    new_atoms = []
                    for idx in best_idx:
                        patch = np.where(self.adjacency[idx] != 0)[0]
                        patch = np.append(patch, idx)
                        new_atoms.append(patch)
                    new_atoms = (
                        np.unique(np.concatenate(new_atoms))
                        if new_atoms
                        else np.array([], dtype=int)
                    )
            elif include_patches:
                b_smooth = np.linalg.norm(
                    self.leadfield_smooth_normed.T @ R, axis=1, ord=2
                )
                b_thresh = thresholding(b_smooth, K)
                best_idx = np.where(b_thresh != 0)[0]
                new_atoms = []
                for idx in best_idx:
                    patch = np.where(self.adjacency[idx] != 0)[0]
                    patch = np.append(patch, idx)
                    new_atoms.append(patch)
                new_atoms = (
                    np.unique(np.concatenate(new_atoms))
                    if new_atoms
                    else np.array([], dtype=int)
                )
            else:  # include_singletons only
                b_sparse = np.linalg.norm(self.leadfield_normed.T @ R, axis=1, ord=2)
                b_thresh = thresholding(b_sparse, K)
                new_atoms = np.where(b_thresh != 0)[0]

            omega = np.unique(np.append(omega, new_atoms)).astype(int)

            x_hat = np.zeros((n_dipoles, n_time))
            if len(omega) > 0:
                x_hat[omega] = self.robust_inverse_solution(
                    self.leadfield_original[:, omega], y
                )

            y_hat = self.leadfield_original @ x_hat
            R = y - y_hat

            residuals = np.append(residuals, np.linalg.norm(R))
            x_hats.append(deepcopy(x_hat))

            if len(residuals) > 1 and residuals[-1] > residuals[-2]:
                break

        x_hat = best_index_residual(residuals, x_hats, plot=False)

        if squeeze:
            x_hat = x_hat[:, 0]
        return x_hat



import logging

import mne
import numpy as np
from scipy.linalg import cho_factor, cho_solve

from ..base import InverseOperator, SolverMeta
from .base_beamformer import BaseBeamformer

logger = logging.getLogger(__name__)


class SolverEBB(BaseBeamformer):
    """Empirical Bayesian Beamformer (EBB) for M/EEG inverse problem.

    Iterative empirical-Bayes / ARD-style beamformer that estimates per-source
    variances (hyperparameters) via EM and forms corresponding spatial filters.

    This is equivalent to Champagne with EM updates, using the efficient
    sensor-space (n_channels x n_channels) formulation.

    References
    ----------
    [1] Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for
        MEG/EEG source imaging. NeuroImage, 44(3), 947-966.
    """

    meta = SolverMeta(
        slug="ebb",
        full_name="Empirical Bayesian Beamformer",
        category="Beamformers",
        description=(
            "Iterative empirical-Bayes / ARD-style beamformer that updates per-source "
            "variances (hyperparameters) from the data and forms corresponding spatial "
            "filters."
        ),
        references=[
            "Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for "
            "MEG/EEG source imaging. NeuroImage, 44(3), 947-966.",
        ],
    )

    def __init__(
        self,
        name="Empirical Bayesian Beamformer",
        reduce_rank=True,
        rank="auto",
        **kwargs,
    ):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        weight_norm=True,
        noise_cov: mne.Covariance | None = None,
        alpha="auto",
        max_iter=300,
        tol=1e-6,
        **kwargs,
    ):
        """Compute the EBB inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            Forward model.
        mne_obj : mne.Evoked | mne.Epochs | mne.io.Raw
            The MNE data object.
        weight_norm : bool
            Apply unit-noise-gain weight normalization.
        noise_cov : mne.Covariance | None
            Noise covariance. None uses identity.
        alpha : float | str
            Regularization parameter.
        max_iter : int
            Maximum EM iterations.
        tol : float
            Convergence tolerance for source variance updates.
        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)

        data = self.unpack_data_obj(mne_obj)

        wf = self.prepare_whitened_forward(noise_cov)
        L = wf.G_white.copy()
        sensor_transform = wf.sensor_transform
        n_channels = wf.n_eff
        n_sources = L.shape[1]

        # Whiten data and compute covariance
        data_w = sensor_transform @ data
        data_cov = self.data_covariance(data_w, center=True, ddof=1)

        inverse_operators = []
        self.alphas = self.get_alphas(reference=data_cov)

        I_m = np.identity(n_channels)

        for alpha in self.alphas:
            gamma = np.ones(n_sources)
            Cn = alpha * I_m

            for n_iter in range(max_iter):
                # Model covariance: Sigma_y = L @ diag(gamma) @ L^T + Cn
                # Invert in sensor space: O(m^3) where m = n_channels
                Sigma_y = (L * gamma) @ L.T + Cn
                try:
                    chol = cho_factor(Sigma_y, overwrite_a=False, check_finite=False)
                    # Reuse solves instead of forming explicit inverse.
                    SiL = cho_solve(chol, L, check_finite=False)
                    SiY = cho_solve(chol, data_w, check_finite=False)
                except np.linalg.LinAlgError:
                    Sigma_y_inv = self.robust_inverse(Sigma_y)
                    SiL = Sigma_y_inv @ L
                    SiY = Sigma_y_inv @ data_w

                # Posterior source estimates: mu = gamma * L^T @ Sigma_y^{-1} @ Y
                mu = gamma[:, None] * (L.T @ SiY)

                # Posterior variance (diagonal only):
                # diag(Sigma_s) = gamma - gamma^2 * diag(L^T Sigma_y^{-1} L)
                z_diag = np.sum(L * SiL, axis=0)
                diag_Sigma_s = gamma - gamma**2 * z_diag

                # EM update: gamma_new = diag(Sigma_s) + mean(mu^2, axis=1)
                gamma_new = diag_Sigma_s + np.mean(mu**2, axis=1)
                gamma_new = np.maximum(gamma_new, 1e-12)

                # Convergence check
                rel_change = np.abs(gamma_new - gamma) / (np.abs(gamma) + 1e-10)
                max_change = np.max(rel_change)

                gamma = gamma_new

                if max_change < tol:
                    logger.info(
                        "EBB converged after %d iterations (max change: %.2e)",
                        n_iter + 1,
                        max_change,
                    )
                    break
            else:
                logger.warning(
                    "EBB reached max iterations (%d), max change: %.2e",
                    max_iter,
                    max_change,
                )

            # Final beamformer weights with converged hyperparameters
            Sigma_y = (L * gamma) @ L.T + Cn
            try:
                chol = cho_factor(Sigma_y, overwrite_a=False, check_finite=False)
                SiI = cho_solve(chol, I_m, check_finite=False)
            except np.linalg.LinAlgError:
                SiI = self.robust_inverse(Sigma_y)
            W = (gamma[:, None] * L.T) @ SiI

            # Unit-noise-gain normalization
            if weight_norm:
                # noise_gain_i = W_i @ Cn @ W_i^T = alpha * ||W_i||^2
                noise_gain = alpha * np.sum(W**2, axis=1)
                W /= np.sqrt(np.maximum(noise_gain, 1e-30))[:, None]

            W_raw = W @ sensor_transform
            inverse_operators.append(W_raw)

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]

        return self



from .datasets import BENCHMARK_DATASETS, DatasetConfig, create_datasets
from .depth_tuning import RECOMMENDED_DEPTH_BY_CATEGORY, RECOMMENDED_DEPTH_BY_SOLVER
from .runner import (
    SOLVER_CATEGORIES,
    BenchmarkResult,
    BenchmarkRunner,
    resolve_solvers,
)
from .visualize import visualize_results

__all__ = [
    "BenchmarkRunner",
    "BenchmarkResult",
    "DatasetConfig",
    "BENCHMARK_DATASETS",
    "create_datasets",
    "RECOMMENDED_DEPTH_BY_CATEGORY",
    "RECOMMENDED_DEPTH_BY_SOLVER",
    "SOLVER_CATEGORIES",
    "resolve_solvers",
    "visualize_results",
]



import logging
from copy import deepcopy

import mne
import numpy as np

from ..base import BaseSolver, SolverMeta
from .utils import soft_threshold

logger = logging.getLogger(__name__)


class SolverMinimumL1NormGPT(BaseSolver):
    """Class for the Minimum Current Estimate inverse solution using
        interesting code from the Chat GPT AI by openai.com (GPT-solver).

        I (Lukas Hecker) prompted the task to write a sparsified eLORETA-type
        inverse solution and this came up with little adjustments required.

        I can't express how weird it is for me, too.

    References
    ----------
    Open AI chat GPT (openai.com)

    """

    meta = SolverMeta(
        acronym="GPT",
        full_name="GPT-derived Sparse Solver",
        category="Experimental",
        description=(
            "Experimental sparse inverse solver derived from an LLM-generated "
            "prototype and lightly adapted; not a published algorithm."
        ),
        references=["tbd"],
        internal=True,
    )

    def __init__(self, name="GPT Solver", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """

        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)
        self.inverse_operators = []
        return self

    def apply_inverse_operator(
        self, mne_obj, max_iter=1000, l1_reg=1e-3, tol=1e-2, depth_weighting=0.0
    ) -> mne.SourceEstimate:
        """Apply the inverse operator.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_reg : float
            Controls the spatial L1 regularization
        tol : float
            Tolerance at which convergence is met.
        depth_weighting : float
            Exponent for depth weighting compensation (0 = no weighting, 1 = full compensation).
            Default 0.0 disables depth weighting.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        """
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data

        source_mat = self.solver_wrap(
            data,
            max_iter=max_iter,
            l1_reg=l1_reg,
            tol=tol,
            depth_weighting=depth_weighting,
        )
        stc = self.source_to_object(source_mat)
        return stc

    def solver_wrap(
        self, y_mat, max_iter=1000, l1_reg=1e-3, tol=1e-2, depth_weighting=0.0
    ):
        srcs = []
        for y in y_mat.T:
            srcs.append(
                self.solve(
                    y,
                    max_iter=max_iter,
                    l1_reg=l1_reg,
                    tol=tol,
                    depth_weighting=depth_weighting,
                )
            )
        return np.stack(srcs, axis=1)

    def solve(self, y, l1_reg=1e-3, max_iter=1000, tol=1e-2, depth_weighting=0.0):
        """
        Solves the EEG inverse problem:
            min_x ||y - Ax||_2^2 + l1_reg * ||x||_1 + l2_reg * ||x||_2^2
        using the FISTA algorithm.

        Parameters
        ----------
        y : ndarray, shape (m,)
            EEG measurements.
        l1_reg : float, optional (default: 1e-3)
            L1 regularization strength.
        max_iter : int, optional (default: 1000)
            Maximum number of iterations to run.
        tol : float, optional (default: 1e-6)
            Tolerance for the stopping criteria.
        depth_weighting : float, optional (default: 0.0)
            Exponent for depth weighting to compensate for superficial source bias.

        Returns
        -------
        x : ndarray, shape (n,)
            Estimated CSDs.
        """

        A = deepcopy(self.leadfield)
        if not self.prep_leadfield and depth_weighting > 0:
            leadfield_norms = np.linalg.norm(A, axis=0)
            leadfield_norms = np.maximum(leadfield_norms, 1e-12)
            depth_weights = leadfield_norms**depth_weighting
            A /= leadfield_norms
            A *= depth_weights

        y_scaled = y.copy()
        # Scale to unit norm
        norm_y = np.linalg.norm(y_scaled)
        if norm_y <= 1e-15:
            return np.zeros(A.shape[1], dtype=float)
        y_scaled /= norm_y

        # Compute step size from Lipschitz constant
        L = np.linalg.norm(A, ord=2) ** 2
        if not np.isfinite(L) or L <= 1e-15:
            return np.zeros(A.shape[1], dtype=float)
        lr = 1.0 / L

        # Calculate initial guess
        x0 = np.linalg.pinv(A) @ y_scaled
        # Scale to unit norm
        x0_norm = np.linalg.norm(x0)
        if x0_norm > 1e-15:
            x0 /= x0_norm
        else:
            x0 = np.zeros_like(x0)

        x = x0.copy()

        for _i in range(max_iter):
            x_prev = x.copy()
            # Gradient descent step (ISTA, no momentum)
            x = x - lr * self.grad_f(x, A, y_scaled)
            # Soft thresholding step
            x = soft_threshold(x, l1_reg * lr)

            # Check stopping criteria
            if np.linalg.norm(x) == 0:
                x = x_prev
                break
            if np.linalg.norm(x - x_prev) < tol:
                break
        # Rescale source
        x = x * norm_y

        return x

    @staticmethod
    def grad_f(x, A, y_scaled):
        """Gradient of the smooth part: A.T @ (A @ x - y)"""
        return A.T.dot(A.dot(x) - y_scaled)



"""
Evaluation script for inverse solution methods.

This script evaluates multiple inverse solution algorithms on simulated EEG/MEG data
and generates comparison plots and JSON results.

Main metrics:
- EMD (Earth Mover's Distance): Main focus metric
- Mean Localization Error: Spatial accuracy of source localization
- Spatial Dispersion: Measure of solution blurring
- Average Precision: Detection performance metric

Usage:
    python scripts/eval_all.py   # Run from project root
    # OR
    cd scripts && python eval_all.py

Outputs:
    - evaluation_results.json: Raw evaluation metrics for all solvers
    - evaluation_results_barplot.png: Bar plots comparing solver performance
"""

import json
import os
import sys

# Add the project root to Python path so we can import from 'invert' module
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import mne  # noqa: E402
import numpy as np  # noqa: E402
from matplotlib import pyplot as plt  # noqa: E402
from scipy.spatial.distance import cdist  # noqa: E402

from invert import Solver  # noqa: E402
from invert.evaluate import evaluate_all  # noqa: E402
from invert.forward import create_forward_model, get_info  # noqa: E402
from invert.simulate import SimulationConfig, SimulationGenerator  # noqa: E402
from invert.util import pos_from_forward  # noqa: E402

pp = dict(surface="inflated", hemi="both", verbose=0, cortex="low_contrast")

sampling = "ico3"
info = get_info(kind="biosemi64")
fwd = create_forward_model(info=info, sampling=sampling)
# fwd["sol"]["data"] /= np.linalg.norm(fwd["sol"]["data"], axis=0)
pos = pos_from_forward(fwd)
leadfield = fwd["sol"]["data"]
n_chans, n_dipoles = leadfield.shape

source_model = fwd["src"]
vertices = [source_model[0]["vertno"], source_model[1]["vertno"]]
adjacency = mne.spatial_src_adjacency(fwd["src"], verbose=0)
distance_matrix = cdist(pos, pos)

sim_config = SimulationConfig(
    batch_repetitions=1,
    batch_size=20,
    n_sources=(2, 2),
    n_orders=(2, 2),
    snr_range=(5, 5),
    correlation_mode=None,
    amplitude_range=(1, 1),
    n_timecourses=200,
    n_timepoints=20,
    add_forward_error=False,
    forward_error=0.1,
    inter_source_correlation=0.0,
    diffusion_parameter=0.1,
    beta_range=(1, 1),
    normalize_leadfield=True,
)

gen = SimulationGenerator(fwd, config=sim_config)
x_test, y_test, sim_info = next(gen.generate())

# solver_names = config.all_solvers
solver_names = ["MNE", "eLORETA", "Champagne", "LCMV", "APSE"]

# Store inverse solutions for all samples
print(f"Computing inverse solutions for {x_test.shape[0]} samples...")
inverse_solutions = {}

for solver_name in solver_names:
    print(f"Processing {solver_name}...")
    solver = Solver(solver_name)

    # Store solutions for all samples
    solver_solutions = []

    for i, x_sample in enumerate(x_test):
        print(f"  Sample {i + 1}/{x_test.shape[0]}")
        evoked = mne.EvokedArray(x_sample.T, info)
        solver.make_inverse_operator(fwd, evoked, alpha="auto")
        stc_hat = solver.apply_inverse_operator(evoked)
        solver_solutions.append(stc_hat)

    inverse_solutions[solver_name] = solver_solutions

# Evaluate all solvers across all samples
print("Evaluating solvers...")
evaluation_results = {}

for solver_name in solver_names:
    print(f"Evaluating {solver_name}...")
    # try:
    stc_solutions = inverse_solutions[solver_name]

    # Store metrics for each sample
    sample_metrics = []

    for i, stc_hat in enumerate(stc_solutions):
        print(f"  Evaluating sample {i + 1}/{len(stc_solutions)}")

        # Get the data arrays - assuming stc_hat has a .data attribute
        if hasattr(stc_hat, "data"):
            y_pred = stc_hat.data
        else:
            y_pred = stc_hat

        # Use the corresponding ground truth sample
        y_true = y_test[i].T
        # convert from sparse to dense
        y_true = y_true

        print(f"    True data shape: {y_true.shape}")
        print(f"    Predicted data shape: {y_pred.shape}")

        print(f"    type(y_true): {type(y_true)}")
        print(f"    type(y_pred): {type(y_pred)}")
        print(f"    type(adjacency): {type(adjacency)}")
        print(f"    type(pos): {type(pos)}")

        # Evaluate using the evaluate_all function
        metrics = evaluate_all(
            y_true=y_true,
            y_pred=y_pred,
            adjacency_true=adjacency,
            adjacency_pred=adjacency,  # Same source space
            pos_true=pos,
            pos_pred=pos,  # Same source space
            mode="dle",
            threshold=0.1,
        )

        sample_metrics.append(metrics)
        print(f"    Sample {i + 1} EMD: {metrics['EMD']:.4f}")

    # Store all sample metrics (without aggregation as requested)
    evaluation_results[solver_name] = sample_metrics

    # Calculate mean EMD for summary
    mean_emd = np.mean([m["EMD"] for m in sample_metrics])
    print(f"  {solver_name} Mean EMD: {mean_emd:.4f}")

    # except Exception as e:
    #     print(f"  Error evaluating {solver_name}: {str(e)}")
    #     # Create dummy metrics for all samples to keep the script running
    #     num_samples = len(x_test)
    #     evaluation_results[solver_name] = [{
    #         'Mean_Localization_Error': np.nan,
    #         'EMD': np.nan,
    #         'sd': np.nan,
    #         'average_precision': np.nan
    #     }] * num_samples

# Save results to JSON (all samples, no aggregation)
output_file = "evaluation_results.json"
print(f"\nSaving results to {output_file}")
with open(output_file, "w") as f:
    # Convert numpy types to native Python types for JSON serialization
    json_results = {}
    for solver, sample_metrics_list in evaluation_results.items():
        json_results[solver] = []
        for sample_metrics in sample_metrics_list:
            sample_json = {}
            for metric, value in sample_metrics.items():
                if isinstance(value, np.ndarray):
                    sample_json[metric] = value.tolist()
                elif isinstance(value, (np.integer, np.floating)):
                    sample_json[metric] = float(value)
                else:
                    sample_json[metric] = value
            json_results[solver].append(sample_json)

    json.dump(json_results, f, indent=2)

# Create bar plots for each metric (aggregated across samples)
print("\nCreating plots...")
metrics_to_plot = ["Mean_Localization_Error", "EMD", "sd", "average_precision"]

fig, axes = plt.subplots(2, 2, figsize=(12, 8))
axes = axes.flatten()

for i, metric in enumerate(metrics_to_plot):
    ax = axes[i]

    # Extract and aggregate values for this metric across all samples
    solvers = list(evaluation_results.keys())
    values = []
    errors = []  # Standard errors for error bars

    for solver in solvers:
        sample_metrics = evaluation_results[solver]
        metric_values = [sample[metric] for sample in sample_metrics]

        # Filter out NaN values
        valid_values = [v for v in metric_values if not np.isnan(v)]

        if valid_values:
            mean_value = np.mean(valid_values)
            std_error = np.std(valid_values) / np.sqrt(len(valid_values))
            values.append(mean_value)
            errors.append(std_error)
        else:
            values.append(np.nan)
            errors.append(0)

    # Handle NaN values for plotting
    plot_values = [0 if np.isnan(v) else v for v in values]
    plot_errors = [0 if np.isnan(v) else e for v, e in zip(values, errors, strict=True)]
    colors = ["red" if np.isnan(v) else "steelblue" for v in values]

    # Create bar plot with error bars
    bars = ax.bar(
        solvers, plot_values, yerr=plot_errors, color=colors, capsize=5, alpha=0.8
    )
    ax.set_title(f"{metric} (Mean ± SE)")
    ax.set_ylabel("Value")

    # Rotate x-axis labels for better readability
    ax.tick_params(axis="x", rotation=45)

    # Add value labels on bars
    for bar, value, error in zip(bars, values, errors, strict=True):
        height = bar.get_height()
        if np.isnan(value):
            label = "NaN"
        else:
            label = f"{value:.3f}±{error:.3f}"
        ax.text(
            bar.get_x() + bar.get_width() / 2.0,
            height + error,
            label,
            ha="center",
            va="bottom",
            fontsize=8,
        )

plt.tight_layout()
plt.show()

# Save the plot
plot_filename = "evaluation_results_barplot.png"
plt.savefig(plot_filename, dpi=300, bbox_inches="tight")
print(f"Plot saved as {plot_filename}")

print("\nEvaluation completed!")
print(f"\nResults summary (averaged across {len(x_test)} samples):")
print("=" * 60)
for solver_name in solver_names:
    sample_metrics = evaluation_results[solver_name]

    # Calculate means across samples
    emd_values = [sample["EMD"] for sample in sample_metrics]
    mle_values = [sample["Mean_Localization_Error"] for sample in sample_metrics]

    # Filter out NaN values
    valid_emd = [v for v in emd_values if not np.isnan(v)]
    valid_mle = [v for v in mle_values if not np.isnan(v)]

    if valid_emd and valid_mle:
        mean_emd = np.mean(valid_emd)
        mean_mle = np.mean(valid_mle)
        std_emd = np.std(valid_emd)
        std_mle = np.std(valid_mle)
        print(
            f"{solver_name:<12}: EMD={mean_emd:.4f}±{std_emd:.4f}, MLE={mean_mle:.4f}±{std_mle:.4f}"
        )
    else:
        print(f"{solver_name:<12}: EMD=NaN, MLE=NaN (evaluation failed)")

print("\nFiles created:")
print(f"  - {output_file} (contains all {len(x_test)} samples without aggregation)")
print(f"  - {plot_filename} (shows mean ± standard error)")
print("\nNote: Lower EMD and MLE values indicate better performance.")



from __future__ import annotations

from dataclasses import dataclass

import mne
import numpy as np

from ..base import SolverMeta
from .base_beamformer import BaseBeamformer
from .esmv_mvpure import SolverESMVMVPURE


@dataclass(frozen=True)
class _ContrastParams:
    eps: float = 1e-12
    gamma_max: float = 1.6  # gamma in [1, 1+gamma_max]
    gamma_fscale: float = 0.05
    patchiness_threshold: float = 0.25


class SolverFlexESMVMVPURE(BaseBeamformer):
    """FlexESMV-style adaptive contrast wrapper around ESMV MV-PURE."""

    meta = SolverMeta(
        slug="flex-esmv-mvpure",
        full_name="Flexible-Extent ESMV (MV-PURE)",
        category="Beamformers",
        description=(
            "FlexESMV-style adaptive contrast reweighting applied on top of an "
            "ESMV MV-PURE beamformer core."
        ),
        references=[
            "Lukas Hecker (2025). Unpublished.",
            "Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R. "
            "(2014). Comparison of beamformers for EEG source signal reconstruction. "
            "Biomedical Signal Processing and Control, 14, 175-188.",
            "Jurkowska, J., Dreszer, J., Lewandowska, M., & Piotrowski, T. (2025). "
            "Multi-Source Neural Activity Indices and Spatial Filters for EEG/MEG "
            "Inverse Problem: An Extension to MNE-Python. arXiv preprint arXiv:2509.14118.",
        ],
    )

    def __init__(
        self,
        name: str = "FlexESMV-MVPURE (ESMV-MVPURE+Contrast) Beamformer",
        params: _ContrastParams | None = None,
        *,
        mvp_n_sources: int | str = "auto",
        mvp_rank: int | str = "auto",
        mvp_max_sources: int = 4,
        spectrum_threshold: float = 0.05,
        selection_corr_threshold: float = 0.97,
        source_indices: list[int] | np.ndarray | None = None,
        reduce_rank: bool = True,
        rank: str | int = "auto",
        **kwargs,
    ):
        if params is None:
            params = _ContrastParams()
        self.name = name
        self._cp = params
        self.mvp_n_sources = mvp_n_sources
        self.mvp_rank = mvp_rank
        self.mvp_max_sources = int(mvp_max_sources)
        self.spectrum_threshold = float(spectrum_threshold)
        self.selection_corr_threshold = float(selection_corr_threshold)
        self.source_indices = (
            None if source_indices is None else np.asarray(source_indices, dtype=int)
        )
        self._inner_kwargs = dict(kwargs)
        self._inner_kwargs.pop("verbose", None)
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        # Keep forward/info handling consistent and defer heavy work to apply.
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        if noise_cov is not None:
            self.coerce_noise_cov(noise_cov)
        self._noise_cov = noise_cov
        _ = self.unpack_data_obj(mne_obj)
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj):  # type: ignore[override]
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)

        base = SolverESMVMVPURE(
            mvp_n_sources=self.mvp_n_sources,
            mvp_rank=self.mvp_rank,
            mvp_max_sources=self.mvp_max_sources,
            spectrum_threshold=self.spectrum_threshold,
            selection_corr_threshold=self.selection_corr_threshold,
            source_indices=self.source_indices,
            reduce_rank=self.reduce_rank,
            rank=self.rank,
            verbose=self.verbose,
            **self._inner_kwargs,
        )
        base.make_inverse_operator(
            self.forward, mne_obj, alpha="auto", noise_cov=self._noise_cov
        )
        stc = base.apply_inverse_operator(mne_obj)
        self.selected_sources_ = getattr(base, "selected_sources_", None)
        self.selected_rank_ = getattr(base, "selected_rank_", None)

        y = stc.data
        if y.ndim != 2 or y.size == 0:
            return stc

        cp = self._cp
        p = np.mean(np.abs(y), axis=1)
        pmax = float(np.max(p)) + cp.eps
        patchiness = float(np.mean((p / pmax) > cp.patchiness_threshold))
        gamma = 1.0 + cp.gamma_max * float(np.exp(-patchiness / cp.gamma_fscale))

        rel = np.clip(p / pmax, cp.eps, 1.0)
        scale = rel ** (gamma - 1.0)
        stc.data[:] = y * scale[:, None]
        return stc



"""Run the official invertmeeg leaderboard evaluation.

Uses biosemi32/ico2, 50 samples per dataset, all solver categories plus CovCNN
variants.  Saves compact results to results/release/leaderboard.json.
"""

from pathlib import Path

from invert.benchmark import (
    RECOMMENDED_DEPTH_BY_CATEGORY,
    RECOMMENDED_DEPTH_BY_SOLVER,
    BenchmarkRunner,
)
from invert.forward import create_forward_model, get_info

if __name__ == "__main__":
    info = get_info(kind="biosemi32")
    fwd = create_forward_model(sampling="ico3", info=info)

    runner = BenchmarkRunner(
        fwd,
        info,
        n_samples=50,  # should be 50 for the full leaderboard
        categories=[
            "beamformer",
            "bayesian",
            "minimum_norm",
            "loreta",
            "music",
            "matching_pursuit",
            "other",
        ],
        exclude_solvers=[
            "SESAME",
            "EBB",
            "ReciPSIICOS-Plain",
            "ReciPSIICOS-Whitened",
            "champagne-ar-em",
            "champagne-tem",
            "exhaustive-ml",
            "cmem",  # scipy.optimize.minimize L-BFGS-B crashes with ThreadPoolExecutor
        ],
        # solvers=[
        #     "CovCNN",
        #     "CovCNN-KL",
        #     "CovCNN-KL-FLEXOMP",
        # ],
        # solver_params={
        #     # Script-level ANN speedup: keep global solver defaults unchanged.
        #     "CovCNN": {"epochs": 120, "patience": 40},
        #     "CovCNN-KL": {"epochs": 250, "patience": 80},
        #     "CovCNN-KL-FLEXOMP": {"epochs": 250, "patience": 80},
        # },
        depth_by_category=RECOMMENDED_DEPTH_BY_CATEGORY,
        depth_by_solver=RECOMMENDED_DEPTH_BY_SOLVER,
        random_seed=42,
        n_jobs=-1,
    )
    runner.run()

    out_path = Path("results/release/leaderboard-test.json")
    out_path.parent.mkdir(parents=True, exist_ok=True)
    runner.save(
        out_path,
        compact=True,
        name="invertmeeg Leaderboard",
        description=(
            "Official leaderboard run using biosemi32/ico2 with 50 samples "
            "per dataset across all solver categories."
        ),
    )



"""Cortical surface visualization of MNE SourceEstimate objects using PyVista.

Renders source activations on the FreeSurfer cortical mesh for
publication-ready 3D figures.
"""

from __future__ import annotations

import numpy as np


def _build_mesh(stc, src, time_idx=0):
    """Build a PyVista mesh from source space with stc data mapped on.

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source estimate.
    src : mne.SourceSpaces
        The source space (forward['src']).
    time_idx : int
        Which time index to read from stc.data.

    Returns
    -------
    meshes : list of pyvista.PolyData
        One mesh per hemisphere with 'activation' scalars set.
    """
    import pyvista as pv

    meshes = []
    offset = 0  # running offset into stc.data rows
    for i, hemi in enumerate(src):
        rr = hemi["rr"] * 1e3  # metres -> mm
        tris = hemi["use_tris"]
        n_faces = len(tris)
        faces = np.column_stack([np.full(n_faces, 3), tris]).ravel()

        mesh = pv.PolyData(rr[hemi["vertno"]], faces)

        # Map stc data onto mesh vertices
        n_stc_verts = len(stc.vertices[i])
        act = np.zeros(len(hemi["vertno"]), dtype=float)

        # Build lookup: vertex number -> index in hemi["vertno"]
        vert_to_idx = {v: idx for idx, v in enumerate(hemi["vertno"])}

        for j, sv in enumerate(stc.vertices[i]):
            if sv in vert_to_idx:
                val = stc.data[offset + j]
                if val.ndim >= 1:
                    val = val[time_idx]
                act[vert_to_idx[sv]] = val

        offset += n_stc_verts
        mesh["activation"] = np.abs(act)
        meshes.append(mesh)

    return meshes


def plot_surface(
    stc,
    src,
    time_idx=None,
    threshold=0.2,
    cmap="magma",
    background="white",
    views=None,
    figsize=(1200, 400),
    title=None,
    screenshot_path=None,
    show=True,
):
    """Plot source activations on the cortical surface mesh.

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source estimate returned by a solver.
    src : mne.SourceSpaces
        Source space, typically ``forward['src']``.
    time_idx : int | None
        Time sample index to plot. If None, uses the time of peak activation.
    threshold : float
        Fraction of peak below which activations are hidden (transparent).
    cmap : str
        Colormap for activations.
    background : str
        Background color of the rendering window.
    views : list of str | None
        Camera views to show. Defaults to ``['lateral', 'medial', 'dorsal']``.
        Options: 'lateral', 'medial', 'dorsal', 'ventral', 'anterior',
        'posterior'.
    figsize : tuple
        Window size (width, height) in pixels.
    title : str | None
        Window title.
    screenshot_path : str | None
        If provided, saves a screenshot to this path and returns the image
        array instead of showing the interactive window.
    show : bool
        Whether to display the interactive window (ignored if
        screenshot_path is set).

    Returns
    -------
    plotter : pyvista.Plotter
        The plotter instance.
    """
    import pyvista as pv

    if views is None:
        views = ["lateral", "medial", "dorsal"]

    # Select time point
    data = stc.data
    if data.ndim == 2 and data.shape[1] > 1:
        if time_idx is None:
            time_idx = np.argmax(np.max(np.abs(data), axis=0))
    else:
        time_idx = 0

    meshes = _build_mesh(stc, src, time_idx=time_idx)

    # Camera presets
    camera_positions = {
        "lateral": [(400, 0, 0), (0, 0, 0), (0, 0, 1)],
        "medial": [(-400, 0, 0), (0, 0, 0), (0, 0, 1)],
        "dorsal": [(0, 0, 400), (0, 0, 0), (0, 1, 0)],
        "ventral": [(0, 0, -400), (0, 0, 0), (0, 1, 0)],
        "anterior": [(0, 400, 0), (0, 0, 0), (0, 0, 1)],
        "posterior": [(0, -400, 0), (0, 0, 0), (0, 0, 1)],
    }

    n_views = len(views)
    shape = (1, n_views)

    off_screen = screenshot_path is not None
    plotter = pv.Plotter(
        shape=shape,
        window_size=figsize,
        off_screen=off_screen,
    )
    plotter.set_background(background)

    # Determine global clim across hemispheres
    all_act = np.concatenate([m["activation"] for m in meshes])
    peak = all_act.max() if all_act.size > 0 else 0
    clim = [threshold * peak, peak] if peak > 0 else [0, 1]

    for vi, view_name in enumerate(views):
        plotter.subplot(0, vi)
        for mesh in meshes:
            # Add the base brain (grey)
            plotter.add_mesh(
                mesh,
                color="lightgrey",
                opacity=1.0,
                smooth_shading=True,
            )
            # Add the activation overlay
            # We use an opacity map: 0 below threshold, 1 above
            # This is a bit complex in PyVista, simpler to just use clim
            # and a colormap that starts with transparency if supported,
            # but standard way is to add mesh twice or use scalars.

            if peak > 0:
                plotter.add_mesh(
                    mesh,
                    scalars="activation",
                    cmap=cmap,
                    clim=clim,
                    show_scalar_bar=False,
                    smooth_shading=True,
                    opacity="linear",  # Fades out low values
                )

        # Add a final scalar bar to the last subplot
        if vi == n_views - 1 and peak > 0:
            plotter.add_scalar_bar(
                title="Activation",
                label_font_size=12,
                title_font_size=14,
                n_labels=5,
            )

        cam = camera_positions.get(view_name)
        if cam is not None:
            plotter.camera_position = cam

        plotter.add_text(
            view_name.capitalize(),
            font_size=12,
            position="upper_left",
            color="black" if background == "white" else "white",
        )

    if title:
        plotter.add_text(title, font_size=16, position="upper_edge")

    if screenshot_path is not None:
        plotter.screenshot(screenshot_path)
        plotter.close()
    elif show:
        plotter.show()

    return plotter



import logging

import mne
import numpy as np

from ..base import BaseSolver, InverseOperator, SolverMeta

logger = logging.getLogger(__name__)


class SolverNLChampagne(BaseSolver):
    """Class for the Noise Learning Champagne (Champagne-NL) inverse solution.

    References
    ----------
    [1] Cai, C., Hashemi, A., Diwakar, M., Haufe, S., Sekihara, K., & Nagarajan,
    S. S. (2021). Robust estimation of noise for electromagnetic brain imaging
    with the champagne algorithm. NeuroImage, 225, 117411.
    """

    meta = SolverMeta(
        acronym="Champagne-NL",
        full_name="Noise-Learning Champagne",
        category="Bayesian",
        description=(
            "Champagne sparse Bayesian learning variant that jointly estimates "
            "per-channel noise parameters/covariance while updating source variances."
        ),
        references=[
            "Cai, C., Hashemi, A., Diwakar, M., Haufe, S., Sekihara, K., & Nagarajan, S. S. (2021). Robust estimation of noise for electromagnetic brain imaging with the champagne algorithm. NeuroImage, 225, 117411.",
        ],
    )

    def __init__(self, name="Champagne-NL", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        max_iter=1000,
        noise_cov: mne.Covariance | None = None,
        prune=True,
        pruning_thresh=1e-3,
        convergence_criterion=1e-8,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum number of iterations.
        noise_cov : [None, numpy.ndarray]
            The noise covariance matrix. Use "None" if not available.
        prune : bool
            If True, the algorithm sets small-activity dipoles to zero
            (pruning).
        pruning_thresh : float
            The threshold at which small gammas (dipole candidates) are set to
            zero.
        convergence_criterion : float
            Minimum change of loss function until convergence is assumed.

        Return
        ------
        self : object returns itself for convenience

        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        if noise_cov is not None:
            self.coerce_noise_cov(noise_cov)
        data = self.unpack_data_obj(mne_obj)

        # SSP projection via standard pipeline (no whitening — NL-Champagne
        # learns its own noise parameters internally).
        wf = self.prepare_whitened_forward(None)
        data_projected = wf.sensor_transform @ np.asarray(data, dtype=float)

        original_leadfield = self.leadfield
        self.leadfield = wf.G_white
        try:
            inverse_operator_projected = self.nl_champagne(
                data_projected,
                alpha=0.01,
                max_iter=max_iter,
                prune=prune,
                pruning_thresh=pruning_thresh,
                convergence_criterion=convergence_criterion,
            )
        finally:
            self.leadfield = original_leadfield

        inverse_operator = inverse_operator_projected @ wf.sensor_transform
        self.inverse_operators = [InverseOperator(inverse_operator, self.name)]
        return self

    def nl_champagne(
        self,
        Y,
        alpha=0.01,
        max_iter=1000,
        prune=True,
        pruning_thresh=1e-3,
        convergence_criterion=1e-8,
    ):
        """Noise Learning Champagne method.

        Parameters
        ----------
        Y : array, shape (n_sensors,)
            measurement vector, capturing sensor measurements
        alpha : float
            The regularization parameter.
        max_iter : int, optional
            The maximum number of inner loop iterations
        prune : bool
            If True, the algorithm sets small-activity dipoles to zero (pruning).
        pruning_thresh : float
            The threshold at which small gammas (dipole candidates) are set to
            zero.
        convergence_criterion : float
            Minimum change of loss function until convergence is assumed.
        Returns
        -------
        x : numpy.ndarray
            Parameter vector, e.g., source vector in the context of BSI (x in the cost
            function formula).

        """
        n_chans, n_dipoles = self.leadfield.shape
        _, n_times = Y.shape
        L = self.leadfield.copy()

        # Scaling of the data (necessary for convergence criterion and pruning
        # threshold)
        Y_scaled = Y.copy()
        Y_scaled /= abs(Y_scaled).mean()

        alpha = np.random.rand(n_dipoles)
        llambda = np.random.rand(n_chans)

        loss_list = []

        for i in range(max_iter):
            previous_alpha = alpha.copy()

            # Sigma_y = L @ diag(alpha) @ L.T + diag(llambda)
            # Using broadcasting: (L * alpha) @ L.T avoids d×d matrix
            Sigma_y = (L * alpha) @ L.T + np.diag(llambda)
            Sigma_y_inv = np.linalg.inv(Sigma_y)

            # 1) Alpha update
            # s_bar = diag(alpha) @ L.T @ Sigma_y_inv @ Y_scaled
            SiL = Sigma_y_inv @ L  # (m, d), compute once
            SiY = Sigma_y_inv @ Y_scaled  # (m, t)
            s_bar = alpha[:, None] * (L.T @ SiY)  # (d, t)

            # z_hat = diag(L.T @ Sigma_y_inv @ L) via column-wise dot
            z_hat = np.sum(L * SiL, axis=0)  # (d,)
            C_s_bar = np.sum(s_bar**2, axis=1)  # (d,)

            alpha = np.sqrt(C_s_bar / (z_hat + 1e-20))

            # 2) LLambda update
            Y_hat = L @ s_bar

            # Convex Bound update
            llambda = np.sqrt(
                np.sum((Y_scaled - Y_hat) ** 2, axis=1) / (np.diag(Sigma_y_inv) + 1e-20)
            )

            # 3) Check convergence
            with np.errstate(divide="ignore", over="ignore", invalid="ignore"):
                sign, log_det_Sigma_y = np.linalg.slogdet(Sigma_y)
            if sign <= 0 or not np.isfinite(log_det_Sigma_y):
                log_det_Sigma_y = np.finfo(float).max / 2
            # tr(Sigma_y_inv @ Y @ Y.T) / n_times
            summation = float(np.sum(SiY * Y_scaled)) / n_times
            loss = log_det_Sigma_y + summation

            loss_list.append(loss)
            if self.verbose > 1:
                logger.debug(
                    f"iter {i}: loss {loss:.2f} ({log_det_Sigma_y:.2f} + {summation / n_times:.2f})"
                )

            if (
                loss == float("-inf")
                or loss == float("inf")
                or np.linalg.norm(alpha) == 0
            ):
                alpha = previous_alpha
                break

            # Check convergence only after first iteration
            if len(loss_list) > 1:
                change = abs(1 - (loss_list[-1] / loss_list[-2]))
                if change < convergence_criterion:
                    if self.verbose > 0:
                        logger.info("Converged!")
                    break

            if prune:
                prune_candidates = alpha < (pruning_thresh * alpha.max())
                alpha[prune_candidates] = 0

        # Final inverse: diag(alpha) @ L.T @ Sigma_y_inv
        Sigma_y = (L * alpha) @ L.T + np.diag(llambda)
        Sigma_y_inv = np.linalg.inv(Sigma_y)
        inverse_operator = alpha[:, None] * (L.T @ Sigma_y_inv)

        return inverse_operator



import importlib.util

import numpy as np
import pytest


def test_prepare_training_data_alignment():
    from invert.adapters.context_lstm import prepare_training_data

    stc = np.array(
        [
            [
                [1.0, 2.0, 3.0, 4.0, 5.0],
                [10.0, 20.0, 30.0, 40.0, 50.0],
            ]
        ]
    )
    x, y = prepare_training_data(stc, lstm_look_back=2)

    assert x.shape == (3, 2, 2)
    assert y.shape == (3, 2)

    np.testing.assert_allclose(x[0], [[1.0, 2.0], [10.0, 20.0]])
    np.testing.assert_allclose(y[0], [3.0, 30.0])

    np.testing.assert_allclose(x[2], [[3.0, 4.0], [30.0, 40.0]])
    np.testing.assert_allclose(y[2], [5.0, 50.0])


def test_standardize_2_is_zero_safe():
    from invert.adapters.context_lstm import standardize_2

    mat = np.array(
        [
            [0.0, 1.0, 0.0],
            [0.0, -2.0, 0.0],
        ]
    )
    scaled = standardize_2(mat)

    assert np.all(np.isfinite(scaled))
    np.testing.assert_allclose(scaled[:, 0], 0.0, atol=0.0)
    np.testing.assert_allclose(scaled[:, 2], 0.0, atol=0.0)
    np.testing.assert_allclose(scaled[:, 1], [0.5, -1.0], atol=1e-12)


def test_contextualize_requires_torch_when_missing():
    if importlib.util.find_spec("torch") is not None:
        pytest.skip("torch is installed")

    from invert.adapters.context_lstm import contextualize, contextualize_bd

    forward = {"sol": {"data": np.zeros((1, 1))}}

    with pytest.raises(ImportError):
        contextualize(None, forward)
    with pytest.raises(ImportError):
        contextualize_bd(None, forward)



from copy import deepcopy

import mne
import numpy as np

from ..base import InverseOperator, SolverMeta
from .base_beamformer import BaseBeamformer
from .utils import build_covariance_candidates


class SolverHOCMCMV(BaseBeamformer):
    """Class for the Higher-Order Covariance Multiple Constrained Minimum Variance (HOCMCMV)
        Beamformer inverse solution [1].

    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    """

    meta = SolverMeta(
        slug="hocmcmv",
        full_name="Higher-Order Covariance Multiple Constrained Minimum Variance",
        category="Beamformers",
        description=(
            "Multiple-constrained minimum-variance beamformer using higher-order "
            "covariance statistics (as implemented here)."
        ),
        references=[
            "Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R. "
            "(2014). Comparison of beamformers for EEG source signal reconstruction. "
            "Biomedical Signal Processing and Control, 14, 175-188.",
        ],
    )

    def __init__(
        self, name="HOCMCMV Beamformer", reduce_rank=True, rank="auto", **kwargs
    ):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        weight_norm=True,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        cov_reg: str = "oas",
        cov_reg_beta: float = 0.05,
        cov_reg_cond_target: float = 1e4,
        order=3,
        verbose=0,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        order : int
            The order of the covariance matrix. Should be a positive integer not
            evenly divisible by two {3, 5, 7, ...}

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        wf = self.prepare_whitened_forward(noise_cov)
        data = self.unpack_data_obj(mne_obj)

        leadfield = wf.G_white
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape

        self.weight_norm = weight_norm

        y = wf.sensor_transform @ data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        y -= y.mean(axis=1, keepdims=True)
        C = self.data_covariance(y, center=False, ddof=1)
        cov_mats, self.alphas, cov_meta = build_covariance_candidates(
            C=C,
            I=I,
            alpha=self.alpha,
            get_alphas_fn=self.get_alphas,
            n_samples=int(y.shape[1]),
            cov_reg=cov_reg,
            cov_reg_beta=float(cov_reg_beta),
            cov_reg_cond_target=float(cov_reg_cond_target),
        )
        if "oas_shrinkage" in cov_meta:
            self._cov_reg_oas_shrinkage = float(cov_meta["oas_shrinkage"])

        inverse_operators = []
        for cov_mat in cov_mats:
            C_inv = self.robust_inverse(cov_mat)
            C_inv_n = self.compute_matrix_power_robust(deepcopy(C_inv), order)

            upper = C_inv @ leadfield
            lower = np.sqrt(
                abs(np.einsum("ij,jk,ki->i", leadfield.T, C_inv_n, leadfield))
            )
            W = upper / lower

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T @ wf.sensor_transform
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]
        return self



import mne
import numpy as np
from scipy.spatial.distance import cdist

from ...util.util import pos_from_forward
from ..base import InverseOperator, SolverMeta
from .base_beamformer import BaseBeamformer
from .utils import build_covariance_candidates


class SolverMCMV(BaseBeamformer):
    """Class for the Multiple Constrained Minimum Variance (MCMV) Beamformer
    inverse solution [1].

    MCMV extends LCMV by applying multiple linear constraints simultaneously.
    This improves robustness in the presence of correlated sources.

    References
    ----------
    [1] Nunes, A. S., Moiseev, A., Kozhemiako, N., Cheung, T., Ribary, U., &
    Doesburg, S. M. (2020). Multiple constrained minimum variance beamformer
    (MCMV) performance in connectivity analyses. NeuroImage, 208, 116386.

    """

    meta = SolverMeta(
        slug="mcmv",
        full_name="Multiple Constrained Minimum Variance",
        category="Beamformers",
        description=(
            "Multiple-constrained extension of LCMV that imposes additional linear "
            "constraints (e.g., to improve robustness with correlated sources)."
        ),
        references=[
            "Nunes, A. S., Moiseev, A., Kozhemiako, N., Cheung, T., Ribary, U., "
            "& Doesburg, S. M. (2020). Multiple constrained minimum variance "
            "beamformer (MCMV) performance in connectivity analyses. "
            "NeuroImage, 208, 116386.",
        ],
    )

    def __init__(self, name="MCMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        weight_norm=True,
        noise_cov: mne.Covariance | None = None,
        alpha="auto",
        cov_reg: str = "oas",
        cov_reg_beta: float = 0.05,
        cov_reg_cond_target: float = 1e4,
        k_constraints=3,
        constraint_mode="distance",
        constraint_indices=None,
        n_preloc_peaks=5,
        verbose=0,
        **kwargs,
    ):
        """Calculate inverse operator using MCMV formula.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        k_constraints : int
            Number of constraints per source. When k=1, reduces to LCMV.
            For k>1, includes k-1 nearest neighbors in constraint matrix.
            Ignored when constraint_mode="lcmv_preloc" (auto-determined)
            or when constraint_indices is provided.
        constraint_mode : str
            How to select constraint sources for each grid point:
            - "distance": k-1 physically nearest sources (Euclidean, default)
            - "similarity": k-1 most correlated leadfield columns
            - "lcmv_preloc": LCMV pre-localization peaks as constraints
            Ignored when constraint_indices is provided.
        constraint_indices : list of array-like, optional
            User-supplied constraint indices, one entry per source. Each
            entry is an array of source indices where the first element is
            the target source and the rest are constraint sources. When
            provided, constraint_mode and k_constraints are ignored.
        n_preloc_peaks : int
            Number of LCMV power peaks to use as constraints (only for
            constraint_mode="lcmv_preloc"). Default 5.

        Return
        ------
        self : object returns itself for convenience

        Notes
        -----
        Implements the MCMV formula: w = C_inv @ G @ inv(G.T @ C_inv @ G) @ f
        where G is the constraint matrix (m x k) and f is the constraint vector.
        """
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        wf = self.prepare_whitened_forward(noise_cov)
        leadfield = wf.G_white.copy()
        sensor_transform = wf.sensor_transform
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape

        self.weight_norm = weight_norm
        self.constraint_mode = constraint_mode

        y = sensor_transform @ data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        y -= y.mean(axis=1, keepdims=True)
        C = self.data_covariance(y, center=False, ddof=1)
        cov_mats, self.alphas, cov_meta = build_covariance_candidates(
            C=C,
            I=I,
            alpha=self.alpha,
            get_alphas_fn=self.get_alphas,
            n_samples=int(y.shape[1]),
            cov_reg=cov_reg,
            cov_reg_beta=float(cov_reg_beta),
            cov_reg_cond_target=float(cov_reg_cond_target),
        )
        if "oas_shrinkage" in cov_meta:
            self._cov_reg_oas_shrinkage = float(cov_meta["oas_shrinkage"])

        # Build per-source constraint index lists
        if constraint_indices is not None:
            constraint_indices_per_source = [
                np.asarray(idx, dtype=int) for idx in constraint_indices
            ]
        else:
            constraint_indices_per_source = self._build_constraint_indices(
                forward=forward,
                leadfield=leadfield,
                C=C,
                cov_mats=cov_mats,
                k_constraints=k_constraints,
                constraint_mode=constraint_mode,
                n_preloc_peaks=n_preloc_peaks,
                n_dipoles=n_dipoles,
            )

        inverse_operators = []
        for cov_mat in cov_mats:
            C_inv = self.robust_inverse(cov_mat)
            W = np.zeros((n_chans, n_dipoles))

            for i in range(n_dipoles):
                indices = constraint_indices_per_source[i]
                k = len(indices)

                if k == 1:
                    G = leadfield[:, i : i + 1]
                    f = np.array([1.0])
                else:
                    G = leadfield[:, indices]
                    f = np.zeros(k)
                    f[0] = 1.0

                # Apply MCMV formula: w = C_inv @ G @ inv(G.T @ C_inv @ G) @ f
                G_T_C_inv = G.T @ C_inv
                G_T_C_inv_G = G_T_C_inv @ G

                # Robust inverse of the k×k matrix
                try:
                    G_T_C_inv_G_inv = np.linalg.inv(G_T_C_inv_G)
                except np.linalg.LinAlgError:
                    G_T_C_inv_G_inv = np.linalg.pinv(G_T_C_inv_G)

                w = C_inv @ G @ G_T_C_inv_G_inv @ f
                W[:, i] = w

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            # Map back to raw sensor space
            inverse_operator = (sensor_transform.T @ W).T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]

        return self

    def _build_constraint_indices(
        self,
        *,
        forward,
        leadfield,
        C,
        cov_mats,
        k_constraints,
        constraint_mode,
        n_preloc_peaks,
        n_dipoles,
    ):
        """Return list of constraint index arrays, one per source.

        Each entry is an array where the first element is the target source
        and the remaining are constraint sources.
        """
        if constraint_mode == "similarity":
            k = min(k_constraints, n_dipoles)
            if k <= 1:
                return [np.array([i]) for i in range(n_dipoles)]
            leadfield_norm = leadfield / (np.linalg.norm(leadfield, axis=0) + 1e-10)
            sim = leadfield_norm.T @ leadfield_norm
            result = []
            for i in range(n_dipoles):
                row = sim[i, :].copy()
                row[i] = -np.inf
                neighbors = np.argsort(row)[::-1][: k - 1]
                result.append(np.concatenate([[i], neighbors]))
            return result

        elif constraint_mode == "distance":
            k = min(k_constraints, n_dipoles)
            if k <= 1:
                return [np.array([i]) for i in range(n_dipoles)]
            pos = pos_from_forward(forward, verbose=0)
            n_sources = len(pos)
            n_orient = n_dipoles // n_sources
            dists = cdist(pos, pos)
            result = []
            for i in range(n_dipoles):
                src_i = i // n_orient
                row = dists[src_i, :].copy()
                row[src_i] = np.inf
                # Find nearest sources, then map back to column indices
                src_neighbors = np.argsort(row)[: k - 1]
                col_neighbors = src_neighbors * n_orient + (i % n_orient)
                result.append(np.concatenate([[i], col_neighbors]))
            return result

        elif constraint_mode == "lcmv_preloc":
            # Compute LCMV power map using the first covariance candidate
            C_reg = cov_mats[0]
            C_inv = self.robust_inverse(C_reg)
            CiL = C_inv @ leadfield  # (n_chans, n_dipoles)

            # power_i = (CiL_i^T C CiL_i) / (l_i^T CiL_i)^2
            numerator = np.einsum("ji,jk,ki->i", CiL, C, CiL)  # CiL^T C CiL diag
            denominator = np.einsum("ji,ji->i", leadfield, CiL)  # l^T CiL diag
            denominator = denominator**2
            denominator = np.maximum(denominator, 1e-30)
            power = numerator / denominator

            # Find local maxima using source-space adjacency
            adjacency = mne.spatial_src_adjacency(forward["src"], verbose=0)
            peak_indices = _find_local_maxima(power, adjacency, n_preloc_peaks)

            # For each source: constraints = [i] + peaks (excluding i)
            result = []
            for i in range(n_dipoles):
                peers = [p for p in peak_indices if p != i]
                result.append(np.array([i] + peers, dtype=int))
            return result

        else:
            raise ValueError(
                f"Unknown constraint_mode={constraint_mode!r}. "
                "Expected 'similarity', 'distance', or 'lcmv_preloc'."
            )


def _find_local_maxima(values, adjacency, n_peaks):
    """Find top local maxima in a source-space scalar map.

    A vertex is a local maximum if its value is >= all adjacent vertices.
    Returns indices of the top ``n_peaks`` local maxima sorted by value.
    """
    from scipy.sparse import issparse

    n = len(values)
    if issparse(adjacency):
        adj_csr = adjacency.tocsr()
    else:
        from scipy.sparse import csr_array

        adj_csr = csr_array(adjacency)
    is_max = np.ones(n, dtype=bool)
    for i in range(n):
        row_start = adj_csr.indptr[i]
        row_end = adj_csr.indptr[i + 1]
        neighbors = adj_csr.indices[row_start:row_end]
        if len(neighbors) > 0 and np.any(values[neighbors] > values[i]):
            is_max[i] = False
    maxima_idx = np.where(is_max)[0]
    # Sort by descending value and take top n_peaks
    order = np.argsort(values[maxima_idx])[::-1]
    return list(maxima_idx[order[:n_peaks]])



from __future__ import annotations

from dataclasses import dataclass

import mne
import numpy as np

from ..base import SolverMeta
from .adapt_flex_esmv import SolverAdaptFlexESMV
from .base_beamformer import BaseBeamformer


@dataclass(frozen=True)
class _ContrastParams:
    eps: float = 1e-12
    gamma_max: float = 1.6  # gamma in [1, 1+gamma_max]
    gamma_fscale: float = 0.05
    patchiness_threshold: float = 0.25


class SolverFlexESMV(BaseBeamformer):
    """FlexESMV: AdaptFlexESMV + adaptive contrast (power-law) to reduce dispersion.

    This is a lightweight, monotone reweighting on the per-dipole mean |y|
    map intended to suppress mid-level leakage (which drives the 50%-FWHM
    dispersion metric) while keeping high-confidence sources intact.
    """

    meta = SolverMeta(
        slug="flex_esmv",
        full_name="Flexible-Extent ESMV",
        category="Beamformers",
        description=(
            "Flex-extent ESMV variant with an adaptive contrast reweighting "
            "postprocess intended to reduce spatial dispersion."
        ),
        references=[
            "Lukas Hecker (2025). Unpublished.",
            "Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R. "
            "(2014). Comparison of beamformers for EEG source signal reconstruction. "
            "Biomedical Signal Processing and Control, 14, 175-188.",
        ],
    )

    def __init__(
        self,
        name: str = "FlexESMV (AdaptFlexESMV+Contrast) Beamformer",
        params: _ContrastParams | None = None,
        reduce_rank: bool = True,
        rank: str | int = "auto",
        **kwargs,
    ):
        if params is None:
            params = _ContrastParams()
        self.name = name
        self._cp = params
        self._inner_kwargs = dict(kwargs)
        self._inner_kwargs.pop("verbose", None)
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(
        self,
        forward,
        mne_obj=None,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        # Just ensure forward/info handling is consistent
        super().make_inverse_operator(forward, mne_obj, *args, alpha=alpha, **kwargs)
        if noise_cov is not None:
            self.coerce_noise_cov(noise_cov)
        self._noise_cov = noise_cov
        _ = self.unpack_data_obj(mne_obj)
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj):  # type: ignore[override]
        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)

        base = SolverAdaptFlexESMV(
            n_orders=2,
            diffusion_parameter=0.1,
            reduce_rank=self.reduce_rank,
            rank=self.rank,
            verbose=self.verbose,
            **self._inner_kwargs,
        )
        base.make_inverse_operator(
            self.forward, mne_obj, alpha="auto", noise_cov=self._noise_cov
        )
        stc = base.apply_inverse_operator(mne_obj)

        y = stc.data
        if y.ndim != 2 or y.size == 0:
            return stc

        cp = self._cp
        p = np.mean(np.abs(y), axis=1)
        pmax = float(np.max(p)) + cp.eps
        patchiness = float(np.mean((p / pmax) > cp.patchiness_threshold))
        gamma = 1.0 + cp.gamma_max * float(np.exp(-patchiness / cp.gamma_fscale))

        rel = np.clip(p / pmax, cp.eps, 1.0)
        scale = rel ** (gamma - 1.0)
        stc.data[:] = y * scale[:, None]
        return stc



import logging

import mne
import numpy as np

from ..base import BaseSolver, SolverMeta

logger = logging.getLogger(__name__)


class SolverMinimumL1L2Norm(BaseSolver):
    """Class for the Minimum L1-L2 Norm solution (MCE) inverse solution. It
        imposes a L1 norm on the source and L2 on the source time courses.

    References
    ----------
    [!] Missing reference - please contact developers if you have it!

    """

    meta = SolverMeta(
        acronym="MxNE",
        full_name="Mixed-Norm Estimate (L1/L2)",
        category="Sparse / Mixed Norm",
        description=(
            "Mixed-norm (L1/L2) inverse that promotes group sparsity across time "
            "by applying an L1 penalty over sources and an L2 norm over each "
            "source time course."
        ),
        references=[
            "Gramfort, A., Kowalski, M., & Hämäläinen, M. (2012). Mixed-norm estimates for the M/EEG inverse problem using accelerated gradient methods. Physics in Medicine and Biology, 57(7), 1937–1961.",
        ],
    )

    def __init__(self, name="Minimum L1-L2 Norm", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha=0.01,
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.prepare_whitened_forward(noise_cov)

        return self

    def apply_inverse_operator(
        self,
        mne_obj,
        alpha="auto",
        max_iter=1000,
        l1_spatial=5e-2,
        l2_temporal=0,
        tol=1e-4,
        depth_weighting=0.0,
        center_data=True,
        scale_l1_by_lmax=True,
    ) -> mne.SourceEstimate:
        """Apply the inverse operator.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_spatial : float
            Controls the spatial L1 regularization
        l2_temporal : float
            Controls additional temporal ridge regularization.
        tol : float
            Tolerance at which convergence is met.
        depth_weighting : float
            Exponent for depth weighting compensation (0 = no weighting, 1 = full compensation).
            Default 0.0 disables depth weighting.
            Use 0 to disable depth weighting entirely.
        center_data : bool
            If True, subtract the per-timepoint channel mean from data.
        scale_l1_by_lmax : bool
            If True, interpret ``l1_spatial`` as a fraction of
            ``lambda_max = max_i ||(A^T Y)_i||_2``.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        """

        data = self.unpack_data_obj(mne_obj)
        self.validate_operator_data_compatibility(data)
        data = self._sensor_transform @ data

        source_mat = self.fista_eeg(
            data,
            alpha=alpha,
            max_iter=max_iter,
            tol=tol,
            l1_spatial=l1_spatial,
            l2_temporal=l2_temporal,
            depth_weighting=depth_weighting,
            center_data=center_data,
            scale_l1_by_lmax=scale_l1_by_lmax,
        )
        stc = self.source_to_object(source_mat)
        return stc

    @staticmethod
    def _group_soft_threshold(X, threshold):
        """Row-wise proximal operator for the L21 norm."""
        row_norms = np.linalg.norm(X, axis=1, keepdims=True)
        row_norms = np.maximum(row_norms, 1e-12)
        scale = np.maximum(1.0 - threshold / row_norms, 0.0)
        return X * scale

    def fista_eeg(
        self,
        y,
        alpha="auto",
        l1_spatial=5e-2,
        l2_temporal=0,
        max_iter=1000,
        tol=1e-4,
        depth_weighting=0.0,
        center_data=True,
        scale_l1_by_lmax=True,
    ):
        """
        Solve a mixed-norm inverse problem over all timepoints jointly:
            min_X 0.5 ||Y - A X||_F^2 + lambda * sum_i ||X_i||_2
                  + 0.5 * l2_temporal * ||X||_F^2

        Parameters:
        - A: array of shape (n_sensors, n_sources)
        - y: array of shape (n_sensors, n_timepoints)
        - l1_spatial: float, mixed-norm regularization strength
        - l2_temporal: float, additional ridge regularization on X
        - max_iter: int, maximum number of iterations
        - tol: float, tolerance for convergence
        - depth_weighting: float, exponent for depth weighting (0=no weighting, 1=full compensation)
                          Default 0.0 disables depth weighting

        Returns:
        - x: array of shape (n_sources, n_timepoints), the solution to the EEG inverse problem
        """
        y_mat = np.asarray(y, dtype=float)
        if y_mat.ndim == 1:
            y_mat = y_mat[:, np.newaxis]

        A = self.leadfield.copy()
        if not self.prep_leadfield and depth_weighting > 0:
            leadfield_norms = np.linalg.norm(A, axis=0)
            leadfield_norms = np.maximum(leadfield_norms, 1e-12)
            depth_weights = leadfield_norms**depth_weighting
            A /= leadfield_norms
            A *= depth_weights

        Y = y_mat.copy()
        if center_data:
            Y -= Y.mean(axis=0, keepdims=True)
        if np.allclose(Y, 0):
            return np.zeros((A.shape[1], Y.shape[1]), dtype=float)

        if alpha != "auto":
            l1_spatial = float(alpha)

        if scale_l1_by_lmax:
            gram_proj = A.T @ Y
            lambda_max = float(np.max(np.linalg.norm(gram_proj, axis=1)))
            lambda_eff = float(l1_spatial) * lambda_max
        else:
            lambda_eff = float(l1_spatial)
        lambda_eff = max(lambda_eff, 0.0)

        L = np.linalg.norm(A, ord=2) ** 2 + float(l2_temporal)
        L = max(float(L), 1e-12)

        x = np.linalg.pinv(A) @ Y
        z = x.copy()
        t = 1.0

        for _i in range(max_iter):
            grad = A.T @ (A @ z - Y) + float(l2_temporal) * z
            v = z - grad / L
            x_new = self._group_soft_threshold(v, lambda_eff / L)
            t_new = (1 + np.sqrt(1 + 4 * t**2)) / 2
            z_new = x_new + (t - 1) / t_new * (x_new - x)
            denom = max(np.linalg.norm(x), 1e-12)
            if np.linalg.norm(x_new - x) / denom < tol:
                break
            x = x_new
            t = t_new
            z = z_new

        return x

    @staticmethod
    def calc_norm(x, n_time):
        return np.sqrt((x**2).sum() / n_time)



from pathlib import Path

from invert.benchmark import BenchmarkRunner
from invert.forward import create_forward_model, get_info

if __name__ == "__main__":
    info = get_info(kind="biosemi32")
    fwd = create_forward_model(sampling="ico2", info=info)

    runner = BenchmarkRunner(
        fwd,
        info,
        n_samples=50,
        categories=[
            "beamformer",
            "bayesian",
            # "minimum_norm",
            # "loreta",
            "music",
            # "matching_pursuit",
            # "other",
        ],
        # solvers=["ESMV", "ESMV-MVPURE", "LCMV", "LCMV-MVPURE", "FlexESMV", "FlexESMV-MVPURE"],
        n_jobs=-1,
    )
    runner.run()
    out_path = Path("results/benchmark_results.json")
    out_path.parent.mkdir(parents=True, exist_ok=True)
    runner.save(out_path, compact=True)



import logging

import mne
import numpy as np
from scipy.sparse.csgraph import laplacian
from scipy.sparse.linalg import eigsh

from invert.util import build_source_adjacency

from ..base import BaseSolver, InverseOperator, SolverMeta

logger = logging.getLogger(__name__)


class SolverSMAP(BaseSolver):
    """Class for the Quadratic regularization and spatial regularization
        (S-MAP) inverse solution [1].

    References
    ----------
    [1] Baillet, S., & Garnero, L. (1997). A Bayesian approach to introducing
    anatomo-functional priors in the EEG/MEG inverse problem. IEEE transactions
    on Biomedical Engineering, 44(5), 374-385.

    """

    meta = SolverMeta(
        acronym="S-MAP",
        full_name="S-MAP (Quadratic + Spatial MAP)",
        category="Minimum Norm",
        description=(
            "MAP-style quadratic inverse with spatial regularization via a "
            "graph Laplacian over the source space."
        ),
        references=[
            "Baillet, S., & Garnero, L. (1997). A Bayesian approach to introducing anatomo-functional priors in the EEG/MEG inverse problem. IEEE Transactions on Biomedical Engineering, 44(5), 374–385.",
        ],
    )

    def __init__(
        self,
        name="S-MAP",
        use_noise_whitener: bool = True,
        use_trace_normalization: bool = True,
        rank_tol: float = 1e-12,
        eps: float = 1e-15,
        **kwargs,
    ):
        self.name = name
        self.use_noise_whitener = bool(use_noise_whitener)
        self.use_trace_normalization = bool(use_trace_normalization)
        self.rank_tol = float(rank_tol)
        self.eps = float(eps)
        super().__init__(**kwargs)
        # The BaseSolver default r-grid tops out at 1e1, which is often too
        # narrow for SMAP's scaled Laplacian penalty. Widen the auto grid.
        self.r_values = np.logspace(-10, 4, int(max(self.n_reg_params, 1)))

    def make_inverse_operator(
        self,
        forward,
        *args,
        alpha="auto",
        noise_cov: mne.Covariance | None = None,
        **kwargs,
    ):
        """Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.

        Return
        ------
        self : object returns itself for convenience
        """
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        if noise_cov is not None:
            self.coerce_noise_cov(noise_cov)

        if self.use_noise_whitener:
            if noise_cov is None:
                noise_cov = self.make_identity_noise_cov(list(self.forward.ch_names))
            wf = self.prepare_whitened_forward(
                noise_cov,
                trace_normalize=self.use_trace_normalization,
                rank_tol=self.rank_tol,
                eps=self.eps,
            )
        else:
            wf = self.prepare_whitened_forward(
                None,
                trace_normalize=self.use_trace_normalization,
                rank_tol=self.rank_tol,
                eps=self.eps,
            )
        if wf.whitener_mode not in ("projected", "none"):
            logger.warning("S-MAP whitener fallback used: %s", wf.whitener_mode)

        leadfield = wf.A if wf.A is not None else wf.G_white
        leadfield_scale = wf.trace_scale
        sensor_transform = wf.sensor_transform

        LTL = leadfield.T @ leadfield

        adjacency = build_source_adjacency(self.forward["src"], verbose=0)
        # First-order Laplacian smoothness: x^T L x (graph Laplacian), rather
        # than squaring the Laplacian as in LORETA-style penalties.
        GTG_sparse = laplacian(adjacency).astype(float)
        GTG = GTG_sparse.toarray()

        if alpha == "auto":
            r_grid = np.asarray(self.r_values, dtype=float)
        else:
            r_grid = np.asarray([float(alpha)], dtype=float)
        max_eig_LTL = float(np.linalg.svd(leadfield, compute_uv=False).max() ** 2)
        try:
            max_eig_penalty = float(
                np.maximum(
                    eigsh(GTG_sparse, k=1, which="LM", return_eigenvectors=False)[0],
                    0.0,
                )
            )
        except Exception:
            max_eig_penalty = float(np.linalg.eigvalsh(GTG).max())
        scale = max_eig_LTL / max(max_eig_penalty, 1e-15)
        self.alphas = list(r_grid * scale)

        inverse_operators = []
        # GG_inv = np.linalg.inv(GTG)
        for alpha in self.alphas:
            kernel_eff = np.linalg.lstsq(
                LTL + float(alpha) * GTG, leadfield.T, rcond=None
            )[0]
            # inverse_operator = GG_inv @ self.leadfield.T @ np.linalg.inv(self.leadfield @ GG_inv @ self.leadfield.T + alpha * np.identity(n_chans))
            inverse_operators.append(
                (float(leadfield_scale) * kernel_eff) @ sensor_transform
            )

        self.inverse_operators = [
            InverseOperator(inverse_operator, self.name)
            for inverse_operator in inverse_operators
        ]

        return self



from .backus_gilbert import SolverBackusGilbert
from .basis_functions import SolverBasisFunctions
from .dspm import SolverDSPM
from .dspm_mne import SolverDSPMMNE
from .eloreta import SolverELORETA
from .epifocus import SolverEPIFOCUS
from .gft_minimum_l1_norm import SolverGFTMinimumL1Norm
from .gft_mne import SolverGFTMNE
from .laura import SolverLAURA
from .loreta import SolverLORETA
from .minimum_l1_l2_norm import SolverMinimumL1L2Norm
from .minimum_l1_norm import SolverMinimumL1Norm
from .minimum_l1_norm_gpt import SolverMinimumL1NormGPT
from .mne import SolverMNE
from .mxne import SolverMxNEBCD
from .self_regularized import SolverSelfRegularizedELORETA
from .sloreta import SolverSLORETA
from .smap import SolverSMAP
from .sslofo import SolverSSLOFO
from .total_variation import SolverTotalVariation
from .utils import calc_eloreta_D2, soft_threshold
from .wmne import SolverWMNE

__all__ = [
    "SolverMNE",
    "SolverGFTMNE",
    "SolverWMNE",
    "SolverDSPM",
    "SolverDSPMMNE",
    "SolverMinimumL1Norm",
    "SolverGFTMinimumL1Norm",
    "SolverMinimumL1NormGPT",
    "SolverMinimumL1L2Norm",
    "SolverMxNEBCD",
    "SolverLORETA",
    "SolverSLORETA",
    "SolverELORETA",
    "SolverSMAP",
    "SolverBasisFunctions",
    "SolverSSLOFO",
    "SolverEPIFOCUS",
    "SolverBackusGilbert",
    "SolverLAURA",
    "SolverSelfRegularizedELORETA",
    "SolverTotalVariation",
    "soft_threshold",
    "calc_eloreta_D2",
]



"""Base class for beamformer solvers.

Beamformers construct spatial filters from covariance structure rather than
fitting the forward model to the data in a residual-minimization sense.
Residual-based regularization selection (GCV, L-curve) is therefore not
applicable; instead, covariance-based strategies (OAS shrinkage, condition
number targeting, trace loading) are used during ``make_inverse_operator()``.
"""

from ..base import BaseSolver


class BaseBeamformer(BaseSolver):
    """Base class for all beamformer solvers.

    Inherits from :class:`BaseSolver` and sets beamformer-specific defaults:

    - ``_is_beamformer = True`` — used by the base class to skip residual-based
      regularization selection and to enable free-orientation support.
    - ``regularisation_method`` defaults to ``"L"`` (L-curve), though for
      beamformers the base class bypasses this entirely and uses index 0.

    Regularization strategy
    -----------------------
    Beamformers optimize spatial filter power (maximize signal-to-noise of a
    virtual sensor), not data-fit residuals. Residual-based regularization
    criteria (GCV, L-curve) are therefore not applicable.

    Instead, regularization is applied to the data covariance during
    ``make_inverse_operator()`` via covariance-based strategies:

    - **OAS shrinkage** — Oracle Approximating Shrinkage toward a scaled
      identity, estimated analytically from the data.
    - **Condition number targeting** — Diagonal loading to reach a target
      condition number (e.g. 100), via ``condition_number_loaded_covariance``.
    - **Trace loading** — ``C + beta * trace(C)/n * I``, controlled by a
      scalar ``beta``.

    The benchmark runner respects this by always using index 0 for beamformers
    rather than running GCV/L-curve selection across multiple inverse operators.
    """

    _is_beamformer = True
    SUPPORTS_VECTOR_ORIENTATION: bool = True

    def __init__(self, *, reduce_rank=True, rank="auto", **kwargs):
        kwargs.setdefault("regularisation_method", "L")
        super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)



import logging

import mne
import numpy as np

logger = logging.getLogger(__name__)

_EPS = 1e-12


def _weighted_minimum_norm(leadfield, data, weights, alpha):
    """Compute weighted minimum-norm estimate for a single timepoint.

    Implements the classic weighted minimum-norm estimator:
        x = W L^T (L W L^T + alpha I)^(-1) y
    where W is diagonal with positive entries given by ``weights``.
    """
    n_chans, n_sources = leadfield.shape
    data = np.asarray(data)
    if data.shape != (n_chans,):
        raise ValueError(
            f"`data` must have shape (n_chans,), got {data.shape} (n_chans={n_chans})."
        )

    weights = np.asarray(weights, dtype=float)
    if weights.shape != (n_sources,):
        raise ValueError(
            f"`weights` must have shape (n_sources,), got {weights.shape} (n_sources={n_sources})."
        )

    weights = np.maximum(weights, _EPS)
    # L @ W, where W is diagonal, can be done by column-wise scaling
    LW = leadfield * weights[np.newaxis, :]
    A = LW @ leadfield.T
    if alpha != 0:
        A = A + alpha * np.eye(n_chans, dtype=A.dtype)

    try:
        z = np.linalg.solve(A, data)
    except np.linalg.LinAlgError:
        z = np.linalg.pinv(A) @ data

    # W @ L^T @ z, where W is diagonal, can be done by element-wise scaling
    return weights * (leadfield.T @ z)


def focuss(stc, evoked, forward, alpha=0.01, max_iter=10, verbose=0):
    """FOCUSS algorithm.

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source Estimate object.
    evoked : mne.EvokedArray
        Evoked EEG data object
    forward : mne.Forward
        The forward model
    alpha : float,
        Regularization parameter
    verbose : int
        Controls verbosity of the program

    Return
    ------
    stc_focuss : mne.SourceEstimate
        The new focussed source estimate
    """

    leadfield = forward["sol"]["data"]
    n_chans, n_dipoles = leadfield.shape
    D = stc.data
    M = evoked.data
    if D.shape[0] != n_dipoles:
        raise ValueError(f"stc has {D.shape[0]} sources but forward has {n_dipoles}.")
    if M.shape[0] != n_chans:
        raise ValueError(f"evoked has {M.shape[0]} channels but forward has {n_chans}.")
    if D.shape[1] != M.shape[1]:
        raise ValueError(
            f"stc and evoked must have the same number of timepoints, got {D.shape[1]} and {M.shape[1]}."
        )

    D_FOCUSS = np.zeros_like(D)
    if verbose:
        logger.info("FOCUSS:")

    for t in range(D.shape[1]):
        if verbose > 0:
            logger.info(f"Time step {t + 1}/{D.shape[1]}")

        x_last = np.asarray(D[:, t], dtype=float)
        weights = np.maximum(np.abs(x_last), _EPS)

        for i in range(max_iter):
            if verbose > 0:
                logger.info(f"Iteration {i + 1}/{max_iter}")

            x_new = _weighted_minimum_norm(leadfield, M[:, t], weights, alpha)

            rel_change = np.linalg.norm(x_new - x_last) / (
                np.linalg.norm(x_last) + _EPS
            )
            x_last = x_new
            weights = np.maximum(np.abs(x_last), _EPS)

            if rel_change < 1e-6:
                if verbose:
                    logger.info(f"Converged at iteration {i + 1}")
                break

        D_FOCUSS[:, t] = x_last

    stc_focuss = stc.copy()
    stc_focuss.data = D_FOCUSS
    return stc_focuss


def s_focuss(stc, evoked, forward, alpha=0.01, percentile=0.01, max_iter=10, verbose=0):
    """Shrinking FOCUSS algorithm. Based on Grech et al. (2008)

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source Estimate object.
    evoked : mne.EvokedArray
        Evoked EEG data object
    forward : mne.Forward
        The forward model
    alpha : float,
        Regularization parameter
    verbose : int
        Controls verbosity of the program

    Return
    ------
    stc_focuss : mne.SourceEstimate
        The new focussed source estimate

    References
    ----------
    [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.

    """

    leadfield = forward["sol"]["data"]
    adjacency = mne.spatial_src_adjacency(forward["src"], verbose=verbose)
    adjacency = (
        adjacency.toarray() if hasattr(adjacency, "toarray") else np.asarray(adjacency)
    )
    n_chans, n_dipoles = leadfield.shape
    D = stc.data
    M = evoked.data

    if D.shape[0] != n_dipoles:
        raise ValueError(f"stc has {D.shape[0]} sources but forward has {n_dipoles}.")
    if M.shape[0] != n_chans:
        raise ValueError(f"evoked has {M.shape[0]} channels but forward has {n_chans}.")
    if D.shape[1] != M.shape[1]:
        raise ValueError(
            f"stc and evoked must have the same number of timepoints, got {D.shape[1]} and {M.shape[1]}."
        )

    D_FOCUSS = np.zeros_like(D)
    if verbose:
        logger.info("Shrinking FOCUSS:")

    for t in range(D.shape[1]):
        if verbose > 0:
            logger.info(f"Time step {t + 1}/{D.shape[1]}")

        x_last = np.asarray(D[:, t], dtype=float)
        weights = np.maximum(np.abs(x_last), _EPS)

        do_smoothing = True
        prev_sparsity = np.inf

        for i in range(max_iter):
            if verbose > 0:
                logger.info(f"Iteration {i + 1}/{max_iter}")

            x_new = _weighted_minimum_norm(leadfield, M[:, t], weights, alpha)

            if do_smoothing:
                x_smooth, sparsity = smooth(x_new, adjacency, percentile=percentile)
                if sparsity < n_chans or sparsity > prev_sparsity:
                    do_smoothing = False
                else:
                    prev_sparsity = sparsity
                    x_new = x_smooth

            rel_change = np.linalg.norm(x_new - x_last) / (
                np.linalg.norm(x_last) + _EPS
            )
            x_last = x_new
            weights = np.maximum(np.abs(x_last), _EPS)

            if rel_change < 1e-6:
                if verbose:
                    logger.info(f"Converged at iteration {i + 1}")
                break

        D_FOCUSS[:, t] = x_last

    stc_focuss = stc.copy()
    stc_focuss.data = D_FOCUSS
    return stc_focuss


def smooth(D_FOCUSS_t, adjacency, percentile=0.01):
    D_FOCUSS_t = np.asarray(D_FOCUSS_t)
    is_column_vector = D_FOCUSS_t.ndim == 2 and D_FOCUSS_t.shape[1] == 1
    x = D_FOCUSS_t[:, 0] if is_column_vector else D_FOCUSS_t.reshape(-1)

    adjacency = (
        adjacency.toarray() if hasattr(adjacency, "toarray") else np.asarray(adjacency)
    )
    n_dipoles = adjacency.shape[0]
    if x.shape[0] != n_dipoles:
        raise ValueError(
            f"adjacency has {n_dipoles} nodes but input has {x.shape[0]} entries."
        )

    max_val = np.abs(x).max()
    if max_val == 0:
        x_out = x.copy()
        return (x_out[:, np.newaxis] if is_column_vector else x_out), 0

    prominent_idc = np.where(np.abs(x) > max_val * percentile)[0]
    if len(prominent_idc) == 0:
        prominent_idc = np.array([int(np.argmax(np.abs(x)))], dtype=int)

    # Keep prominent sources and their neighbors. Note that MNE adjacency often has a
    # zero diagonal, so we add prominent indices explicitly.
    neighbor_mask = adjacency[prominent_idc, :].sum(axis=0) > 0
    neighbor_mask[prominent_idc] = True
    neighbor_cat = np.where(neighbor_mask)[0]

    x_smoothed = np.zeros_like(x)
    for idx in neighbor_cat:
        neighbor_idc = np.where(adjacency[idx, :] != 0)[0]
        if not np.any(neighbor_idc == idx):
            neighbor_idc = np.append(neighbor_idc, idx)
        x_smoothed[idx] = np.mean(x[neighbor_idc])

    sparsity = len(prominent_idc)
    return (x_smoothed[:, np.newaxis] if is_column_vector else x_smoothed), sparsity
